# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from copy import deepcopy

import jax
import pandas as pd
import pytest

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect
from jaxley.modules.base import View
from jaxley.synapses import TestSynapse
from jaxley.utils.cell_utils import loc_of_index, local_index_of_loc
from jaxley.utils.misc_utils import cumsum_leading_zero


def test_getitem(SimpleBranch, SimpleCell, SimpleNet):
    branch = SimpleBranch(4)
    cell = SimpleCell(3, 4)
    net = SimpleNet(3, 3, 4)

    # test API equivalence
    assert all(net.cell(0).branch(0).show() == net[0, 0].show())
    assert all(net.cell([0, 1]).branch(0).show() == net[[0, 1], 0].show())
    assert all(net.cell(0).branch([0, 1]).show() == net[0, [0, 1]].show())
    assert all(net.cell([0, 1]).branch([0, 1]).show() == net[[0, 1], [0, 1]].show())

    assert all(net.cell(0).branch(0).comp(0).show() == net[0, 0, 0].show())
    assert all(cell.branch(0).comp(0).show() == cell[0, 0].show())
    assert all(cell.branch(0).show() == cell[0].show())

    # test indexing of comps
    assert branch[:2]
    assert cell[:2, :2]
    assert net[:2, :2, :2]

    # test iterability
    for cell in net.cells:
        pass

    for cell in net.cells:
        for branch in cell.branches:
            for comp in branch.comps:
                pass

    for comp in net[0, 0].comps:
        pass


def test_loc_v_comp(SimpleBranch):
    branch = SimpleBranch(4)
    ncomps = branch.ncomp_per_branch
    branch_ind = 0

    assert np.all(branch.comp(0).show() == branch.loc(0.0).show())
    assert np.all(branch.comp(3).show() == branch.loc(1.0).show())

    inferred_loc = loc_of_index(2, branch_ind, ncomps)
    assert np.all(branch.loc(inferred_loc).show() == branch.comp(2).show())

    inferred_ind = local_index_of_loc(0.4, branch_ind, ncomps)
    assert np.all(branch.comp(inferred_ind).show() == branch.loc(0.4).show())


def test_shape(SimpleComp, SimpleBranch, SimpleCell, SimpleNet):
    comp = SimpleComp()
    branch = SimpleBranch(4)
    cell = SimpleCell(3, 4)
    net = SimpleNet(3, 3, 4)

    assert net.shape == (3, 3 * 3, 3 * 3 * 4)
    assert cell.shape == (3, 3 * 4)
    assert branch.shape == (4,)
    assert comp.shape == ()

    assert net.cell("all").shape == net.shape
    assert cell.branch("all").shape == cell.shape

    assert net.cell("all").shape == (3, 3 * 3, 3 * 3 * 4)
    assert net.cell("all").branch("all").shape == (3, 3 * 3, 3 * 3 * 4)
    assert net.cell("all").branch("all").comp("all").shape == (3, 3 * 3, 3 * 3 * 4)

    assert net.cell(0).shape == (1, 3, 3 * 4)
    assert net.cell(0).branch(0).shape == (1, 1, 4)
    assert net.cell(0).branch(0).comp(0).shape == (1, 1, 1)


def test_set_and_insert(SimpleBranch, SimpleCell, SimpleNet):
    branch = SimpleBranch(4)
    cell = SimpleCell(5, 4)
    net = SimpleNet(5, 5, 4)
    net1 = deepcopy(net)
    net2 = deepcopy(net)
    net3 = deepcopy(net)
    net4 = deepcopy(net)

    # insert multiple
    net1.cell([0, 1]).branch(0).insert(HH())
    net1.cell(1).branch([0, 1]).insert(HH())
    net1.cell([2, 3]).branch([2, 3]).insert(HH())
    net1.cell(4).branch(4).comp(0).insert(HH())

    net2[[0, 1], 0].insert(HH())
    net2[1, [0, 1]].insert(HH())
    net2[[2, 3], [2, 3]].insert(HH())
    net2[4, 4, 0].insert(HH())

    # set multiple
    net1.cell([0, 1]).branch(0).set("length", 2.0)
    net1.cell(1).branch([0, 1]).set("length", 2.0)
    net1.cell([2, 3]).branch([2, 3]).set("length", 2.0)
    net1.cell(4).branch(4).comp(0).set("length", 2.0)

    net2[[0, 1], 0].set("length", 2.0)
    net2[1, [0, 1]].set("length", 2.0)
    net2[[2, 3], [2, 3]].set("length", 2.0)
    net2[4, 4, 0].set("length", 2.0)

    # insert / set at different levels
    net3.insert(HH())  # insert at net level
    net3.cell(0).insert(HH())  # insert at cell level
    net3.cell(2).branch(2).insert(HH())  # insert at cell level
    net3.cell(4).branch(3).comp(0).insert(HH())  # insert at cell level

    net3.set("length", 2.0)
    net3.cell(0).set("length", 2.0)
    net3.cell(2).branch(2).set("length", 2.0)
    net3.cell(4).branch(3).comp(0).set("length", 2.0)

    net4.insert(HH())
    net4.cell(0).insert(HH())
    net4.cell(2).branch(2).insert(HH())
    net4.cell(4).branch(3).comp(0).insert(HH())

    net4.set("length", 2.0)
    net4.cell(0).set("length", 2.0)
    net4.cell(2).branch(2).set("length", 2.0)
    net4.cell(4).branch(3).comp(0).set("length", 2.0)

    assert all(net1.show() == net2.show())
    assert all(net3.show() == net4.show())

    # insert at into a branch
    branch1 = deepcopy(branch)
    branch2 = deepcopy(branch)

    branch1.comp(0).insert(HH())
    branch2[0].insert(HH())
    assert all(branch1.show() == branch2.show())

    # test insert multiple stimuli
    single_current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    batch_of_currents = np.vstack([single_current for _ in range(4)])

    cell1 = deepcopy(cell)
    cell2 = deepcopy(cell)

    cell1.branch(0).stimulate(single_current)
    cell1.branch(1).comp(0).stimulate(single_current)
    cell1.branch(0).stimulate(batch_of_currents)
    cell1.branch(0).record("v")

    cell2[0].stimulate(single_current)
    cell2[1].comp(0).stimulate(single_current)
    cell2[0].stimulate(batch_of_currents)
    cell2.branch(0).record("v")

    assert np.all(cell1.externals["i"] == cell2.externals["i"])
    assert np.all(cell1.external_inds["i"] == cell2.external_inds["i"])
    assert np.all(cell1.recordings == cell2.recordings)


def test_local_indexing(SimpleNet):
    net = SimpleNet(2, 5, 4)

    local_idxs = net.nodes[
        ["local_cell_index", "local_branch_index", "local_comp_index"]
    ]
    idx_cols = ["global_cell_index", "global_branch_index", "global_comp_index"]
    global_index = 0
    for cell_idx in range(2):
        for branch_idx in range(5):
            for comp_idx in range(4):
                assert np.all(
                    local_idxs.iloc[global_index] == [cell_idx, branch_idx, comp_idx]
                )
                global_index += 1


def test_indexing_a_compartment_of_many_branches(SimpleBranch):
    branch1 = SimpleBranch(ncomp=3)
    branch2 = SimpleBranch(ncomp=4)
    branch3 = SimpleBranch(ncomp=5)
    cell1 = jx.Cell([branch1, branch2, branch3], parents=[-1, 0, 0])
    cell2 = jx.Cell([branch3, branch2], parents=[-1, 0])
    net = jx.Network([cell1, cell2])

    # Indexing a single compartment of multiple branches is not supported with `loc`.
    # TODO FROM #447: Reevaluate what kind of indexing is allowed and which is not!
    # with pytest.raises(NotImplementedError):
    #     net.cell("all").branch("all").loc(0.0)
    # with pytest.raises(NotImplementedError):
    #     net.cell(0).branch("all").loc(0.0)
    # with pytest.raises(NotImplementedError):
    #     net.cell("all").branch(0).loc(0.0)

    # Indexing a single compartment of multiple branches is still supported with `comp`.
    net.cell("all").branch("all").comp(0)
    net.cell(0).branch("all").comp(0)
    net.cell("all").branch(0).comp(0)

    # Indexing many single compartment of multiple branches is always supported.
    net.cell("all").branch("all").loc("all")
    net.cell(0).branch("all").loc("all")
    net.cell("all").branch(0).loc("all")


# make sure all attrs in module also have a corresponding attr in view
def test_view_attrs(SimpleComp, SimpleBranch, SimpleCell, SimpleNet):
    """Check if all attributes of Module have a corresponding attribute in View.

    To ensure that View behaves like a Module as much as possible, View should support
    all attributes of Module. This test checks if all attributes of Module have a
    corresponding attribute in View. Also checks if the types of the attributes match.
    """
    # attributes of Module that do not have to exist in View
    exceptions = ["view"]

    # TODO: Types are inconsistent between different Modules
    exceptions += ["_cumsum_nbranches"]

    # TODO FROM #447: should be added to View in the future
    exceptions += [
        "_internal_node_inds",
        "_par_inds",
        "_child_inds",
        "_child_belongs_to_branchpoint",
        "_solve_indexer",
        "_comp_edges",
        "_n_nodes",
        "_data_inds",
        "_indices_jax_spsolve",
        "_indptr_jax_spsolve",
        "_off_diagonal_inds",
        "_n_nodes",
        "_data_inds",
        "_indices_jax_spsolve",
        "_indptr_jax_spsolve",
        "_off_diagonal_inds",
        "_dhs_solve_indexer",
        "_exp_euler_solve_indexer",
        "_solver_device",
    ]  # for base/comp
    exceptions += ["comb_children"]  # for cell
    exceptions += [
        "_cells_list",
        "_cumsum_nbranchpoints_per_cell",
        "_cumsum_ncomp_per_cell",
    ]  # for network

    for module in [
        SimpleComp(),
        SimpleBranch(2),
        SimpleCell(2, 3),
        SimpleNet(2, 2, 3, connect=True),
    ]:
        for name, attr in module.__dict__.items():
            if name not in exceptions:
                # check if attr is in view
                view = View(module)
                assert hasattr(view, name), f"View missing attribute: {name}"
                # check if types match
                assert type(getattr(module, name)) == type(
                    getattr(view, name)
                ), f"Type mismatch: {name}, Module type: {type(getattr(module, name))}, View type: {type(getattr(view, name))}"


def test_view_supported_index_types(SimpleComp, SimpleBranch, SimpleCell, SimpleNet):
    """Check if different ways to index into Modules/Views work correctly."""
    # test int, range, slice, list, np.array, pd.Index

    for module in [
        SimpleComp(),
        SimpleBranch(4),
        SimpleCell(3, 4),
        SimpleNet(2, 3, 4),
    ]:
        index_types = [
            0,
            range(3),
            slice(0, 3),
            [0, 1, 2],
            np.array([0, 1, 2]),
            np.array([0, 1, 2], dtype=np.int32),
            np.array([0, 1, 2], dtype=np.int64),
            pd.Index([0, 1, 2]),
            pd.Index([0, 1, 2]).to_numpy(),
            np.array([True, False, True, False] * 100)[: len(module.nodes)],
        ]

        # comp.comp is not allowed
        all_inds = module.nodes.index.to_numpy()
        if not isinstance(module, jx.Compartment):
            # `_reformat_index` should always return a np.ndarray
            for index in index_types:
                assert isinstance(
                    module._reformat_index(index), np.ndarray
                ), f"Failed for {type(index)}"

                # test indexing into module and view
                assert module.comp(index), f"Failed for {type(index)}"
                assert View(module).comp(index), f"Failed for {type(index)}"

                expected_inds = all_inds[index]
                assert np.all(module.select(nodes=index).nodes.index == expected_inds)

            # for loc test float and list of floats
            assert module.loc(0.0), "Failed for float"
            assert module.loc([0.0, 0.5, 1.0]), "Failed for List[float]"
        else:
            with pytest.raises(AssertionError):
                module.comp(0)

        if isinstance(module, jx.Network):
            connect(module[0, 0, :], module[1, 0, :], TestSynapse())
            all_inds = module.edges.index.to_numpy()
            for index in index_types[:-1] + [np.array([True, False, True, False])]:
                expected_inds = all_inds[index]
                assert np.all(module.select(edges=index).edges.index == expected_inds)


def test_select(SimpleNet):
    """Ensure `select` works correctly and returns expected View of Modules."""
    net = SimpleNet(3, 3, 2, connect=False)
    connect(net[0, 0, :], net[1, 0, :], TestSynapse())

    np.random.seed(0)

    # select only nodes
    inds = np.random.choice(net.nodes.index, replace=False, size=5)
    view = net.select(nodes=inds)
    assert np.all(view.nodes.index == inds), "Selecting nodes by index failed"

    # select only edges
    inds = np.random.choice(net.edges.index, replace=False, size=2)
    view = net.select(edges=inds)
    assert np.all(view.edges.index == inds), "Selecting edges by index failed"

    # check if pre and post comps of edges are in nodes
    edge_node_inds = np.unique(
        view.edges[["pre_index", "post_index"]].to_numpy().flatten()
    )
    assert np.all(
        view.nodes.index == edge_node_inds
    ), "Selecting edges did not yield the correct nodes."

    # select nodes and edges
    node_inds = np.random.choice(net.nodes.index, replace=False, size=5)
    edge_inds = np.random.choice(net.edges.index, replace=False, size=2)
    view = net.select(nodes=node_inds, edges=edge_inds)
    assert np.all(
        view.nodes.index == node_inds
    ), "Selecting nodes and edges by index failed for nodes."
    assert np.all(
        view.edges.index == edge_inds
    ), "Selecting nodes and edges by index failed for edges."


def test_viewing(SimpleCell, SimpleNet):
    """Test that the View object is working correctly."""
    cell = SimpleCell(3, 3)
    net = SimpleNet(3, 3, 3)

    # test parameter sharing works correctly
    nodes1 = net.branch(0).comp("all").nodes
    nodes2 = net.branch(0).nodes
    nodes3 = net.cell(0).nodes
    control_params1 = nodes1.pop("controlled_by_param")
    control_params2 = nodes2.pop("controlled_by_param")
    control_params3 = nodes3.pop("controlled_by_param")
    assert np.all(nodes1 == nodes2), "Nodes are not the same"
    assert np.all(
        control_params1 == nodes1["global_comp_index"]
    ), "Parameter sharing is not correct"
    assert np.all(
        control_params2 == nodes2["global_branch_index"]
    ), "Parameter sharing is not correct"
    assert np.all(
        control_params3 == nodes3["global_cell_index"]
    ), "Parameter sharing is not correct"

    # test local and global indexes match the expected targets
    for view, local_targets, global_targets in zip(
        [
            net.branch(0),  # shows every comp on 0th branch of all cells
            cell.branch("all"),  # shows all branches and comps of cell
            net.cell(0).comp(0),  # shows every 0th comp for every branch on 0th cell
            net.comp(0),  # shows 0th comp of every branch of every cell
            cell.comp(0),  # shows 0th comp of every branch of cell
        ],
        [[0, 1, 2] * 3, [0, 1, 2] * 3, [0] * 3, [0] * 9, [0] * 3],
        [
            [0, 1, 2, 9, 10, 11, 18, 19, 20],
            list(range(9)),
            [0, 3, 6],
            list(range(0, 27, 3)),
            list(range(0, 9, 3)),
        ],
    ):
        assert np.all(
            view.nodes["local_comp_index"] == local_targets
        ), "Indices do not match that of the target"
        assert np.all(
            view.nodes["global_comp_index"] == global_targets
        ), "Indices do not match that of the target"

    with pytest.raises(ValueError):
        net.scope("global").comp(999)  # Nothing should be in View


def test_scope(SimpleCell):
    """Ensure scope has the intended effect for Modules and Views."""
    cell = SimpleCell(3, 3)

    view = cell.scope("global").branch(1)
    assert view._scope == "global"
    view = view.scope("local").comp(0)
    assert np.all(
        view.nodes[["global_branch_index", "global_comp_index"]] == [1, 3]
    ), "Expected [1,3] but got {}".format(
        view.nodes[["global_branch_index", "global_comp_index"]]
    )

    cell.set_scope("global")
    assert cell._scope == "global"
    view = cell.branch(1).comp(3)
    assert np.all(
        view.nodes[["global_branch_index", "global_comp_index"]] == [1, 3]
    ), "Expected [1,3] but got {}".format(
        view.nodes[["global_branch_index", "global_comp_index"]]
    )

    cell.set_scope("local")
    assert cell._scope == "local"
    view = cell.branch(1).comp(0)
    assert np.all(
        view.nodes[["global_branch_index", "global_comp_index"]] == [1, 3]
    ), "Expected [1,3] but got {}".format(
        view.nodes[["global_branch_index", "global_comp_index"]]
    )


def test_context_manager(SimpleCell):
    """Test that context manager works correctly for Module."""
    cell = SimpleCell(3, 3)

    with cell.branch(0).comp(0) as comp:
        comp.set("v", -71)
        comp.set("radius", 0.123)

    with cell.branch(1).comp([0, 1]) as comps:
        comps.set("v", -71)
        comps.set("radius", 0.123)

    assert np.all(
        cell.branch(0).comp(1).nodes[["v", "radius"]] == [-70, 1.0]
    ), "Set affected nodes not in context manager View."
    assert np.all(
        cell.branch(0).comp(0).nodes[["v", "radius"]] == [-71, 0.123]
    ), "Context management of View not working."
    assert np.all(
        cell.branch(1).comp([0, 1]).nodes[["v", "radius"]] == [-71, 0.123]
    ), "Context management of View not working."


def test_iter(SimpleBranch):
    """Test that __iter__ works correctly for all modules."""
    branch1 = SimpleBranch(2)
    branch2 = SimpleBranch(3)
    cell = jx.Cell([branch1, branch1, branch2], parents=[-1, 0, 0])
    net = jx.Network([cell] * 2)

    # test iterating over branches with different numbers of compartments
    assert np.all(
        [
            len(branch.nodes) == expected_len
            for branch, expected_len in zip(cell.branches, [2, 2, 3])
        ]
    ), "__iter__ failed for branches with different numbers of compartments."

    # test iterating using cells, branches, and comps properties
    nodes1 = []
    for cell in net.cells:
        for branch in cell.branches:
            for comp in branch.comps:
                nodes1.append(comp.nodes)
    assert len(nodes1) == len(net.nodes), "Some compartments were skipped in iteration."

    nodes2 = []
    for cell in net:
        for branch in cell:
            for comp in branch:
                nodes2.append(comp.nodes)
    assert len(nodes2) == len(net.nodes), "Some compartments were skipped in iteration."
    assert np.all(
        [np.all(n1 == n2) for n1, n2 in zip(nodes1, nodes2)]
    ), "__iter__ is not consistent with [comp.nodes for cell in net.cells for branches in cell.branches for comp in branches.comps]"

    assert np.all(
        [len(comp.nodes) for comp in net[0, 0].comps] == [1, 1]
    ), "Iterator yielded unexpected number of compartments"

    # 0th comp in every branch (3), 1st comp in every branch (3), 2nd comp in (every) branch (only 1 branch with > 2 comps)
    assert np.all(
        [len(comp.nodes) for comp in net[0].comps] == [3, 3, 1]
    ), "Iterator yielded unexpected number of compartments"

    # 0th comp in every branch for every cell (6), 1st comp in every branch for every cell , 2nd comp in (every) branch for every cell
    assert np.all(
        [len(comp.nodes) for comp in net.comps] == [6, 6, 2]
    ), "Iterator yielded unexpected number of compartments"

    for comp in branch1:
        comp.set("v", -72)
    assert np.all(branch1.nodes["v"] == -72), "Setting parameters with __iter__ failed."

    # needs to be redefined because cell was overwritten with View object
    cell = jx.Cell([branch1, branch1, branch2], parents=[-1, 0, 0])
    for branch in cell:
        for comp in branch:
            comp.set("v", -73)
    assert np.all(cell.nodes["v"] == -73), "Setting parameters with __iter__ failed."


def test_synapse_and_channel_filtering(SimpleNet):
    """Test that synapses and channels are filtered correctly by View."""
    net = SimpleNet(3, 3, 3, connect=False)
    net.insert(HH())
    connect(net[0, 0, :], net[1, 0, :], TestSynapse())

    assert np.all(net.cell(0).HH.nodes == net.HH.cell(0).nodes)
    view1 = net.cell([0, 1]).TestSynapse
    nodes1 = view1.nodes
    edges1 = view1.edges
    view2 = net.TestSynapse.cell([0, 1])
    nodes2 = view2.nodes
    edges2 = view2.edges
    nodes_control_param1 = nodes1.pop("controlled_by_param")
    nodes_control_param2 = nodes2.pop("controlled_by_param")
    edges_control_param1 = edges1.pop("controlled_by_param")
    edges_control_param2 = edges2.pop("controlled_by_param")

    # convert to dict so order of cols and index dont matter for __eq__
    assert nodes1.to_dict() == nodes2.to_dict()
    assert np.all(nodes_control_param1 == 0)
    assert np.all(nodes_control_param2 == nodes2["global_cell_index"])

    assert np.all(edges1 == edges2)


def test_view_equals_module(SimpleComp, SimpleBranch):
    """Test that View behaves the same as Module for important attrs and methods."""
    comp = SimpleComp(copy=True)
    branch = SimpleBranch(3)

    comp.insert(HH())
    branch.comp([0, 1]).insert(HH())

    comp.set("v", -71.2)
    branch.comp(0).set("v", -71.2)

    comp.record("v")
    branch.comp([0, 1]).record("v")

    comp.stimulate(np.zeros(100))
    branch.comp([0, 1]).stimulate(np.zeros(100))

    comp.make_trainable("HH_gNa")
    comp.make_trainable("HH_gK")
    branch.comp([0, 1]).make_trainable("HH_gNa")
    branch.make_trainable("HH_gK")

    # test deleting subset of attributes
    branch.comp(1).delete_trainables()
    branch.comp(1).delete_recordings()
    branch.comp(1).delete_stimuli()

    assert (
        branch.comp(1).trainable_params == [] and branch.comp(0).trainable_params != []
    )
    assert branch.comp(1).recordings.empty and not branch.comp(0).recordings.empty
    assert branch.comp(1).externals == {} and branch.comp(0).externals != {}

    # convert to dict so order of cols and index dont matter for __eq__
    assert comp.nodes.to_dict() == branch.comp(0).nodes.to_dict()

    assert comp.trainable_params == branch.comp(0).trainable_params
    assert comp.indices_set_by_trainables == branch.comp(0).indices_set_by_trainables
    assert np.all(comp.recordings == branch.comp(0).recordings)
    assert np.all(
        [
            np.all([np.all(v1 == v2), k1 == k2])
            for (k1, v1), (k2, v2) in zip(
                comp.externals.items(), branch.comp(0).externals.items()
            )
        ]
    )

    assert comp._comp_edges.columns.equals(branch.comp(0)._comp_edges.columns)


@pytest.mark.parametrize("ncomp", [1, 4])
def test_comp_edge_indexing(SimpleCell, ncomp: int):
    """Test whether `_comp_edges` are tracked correctly when viewing."""
    ncomp = 4
    cell = SimpleCell(5, ncomp)
    # branchpoint on one side of branch.
    assert len(cell.branch(0)._comp_edges) == (ncomp - 1) * 2 + 2
    # branchpoint on both sides of branch.
    assert len(cell.branch(1)._comp_edges) == (ncomp - 1) * 2 + 4
    # branchpoint on one side of branch.
    assert len(cell.branch(2)._comp_edges) == (ncomp - 1) * 2 + 2

    assert len(cell.branch([0, 1])._comp_edges) == (ncomp - 1) * 2 * 2 + 2 + 4

    # .comp() should never return any branchpoints.
    assert len(cell.branch(1).comp(0)._comp_edges) == 0
    assert len(cell.branch([0, 1]).comp(0)._comp_edges) == 0
    assert len(cell.branch("all").comp(0)._comp_edges) == 0

    # .loc() should not return any branchpoints if the value != 0.0 or 1.0.
    assert len(cell.branch(1).loc(0.1)._comp_edges) == 0
    assert len(cell.branch([0, 1]).loc(0.2)._comp_edges) == 0
    assert len(cell.branch("all").loc(0.9)._comp_edges) == 0

    # .loc(0.0) or .loc(1.0) should return edges to the branchpoint.
    assert len(cell.branch(1).loc(0.0)._comp_edges) == 2
    assert len(cell.branch(0).loc(0.0)._comp_edges) == 0  # 0 is a tip branch.
    assert len(cell.branch(1).loc(1.0)._comp_edges) == 2
    assert len(cell.branch([0, 1]).loc(1.0)._comp_edges) == 4  # 0 + 1 have endpoints.
    assert len(cell.branch([0, 2]).loc(1.0)._comp_edges) == 2  # only 0 has endpoint.

    # Finally, a few checks that `.loc(0.0)` and `.loc(1.0)` return the correct
    # `comp_edges`.
    assert ncomp - 1 in cell.branch(0).loc(1.0)._comp_edges["sink"].to_numpy().tolist()
    assert ncomp in cell.branch(1).loc(0.0)._comp_edges["sink"].to_numpy().tolist()
    assert (
        2 * ncomp - 1
        not in cell.branch(1).loc(0.0)._comp_edges["sink"].to_numpy().tolist()
    )


def test_module_inheritance():
    """Test inheritance of modules works properly. (see #590)"""

    class CustomCompartment(jx.Compartment):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    class CustomBranch(jx.Branch):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    class CustomCell(jx.Cell):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    class CustomNetwork(jx.Network):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    class CustomChildNetwork(CustomNetwork):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    custom_comp = CustomCompartment()
    custom_branch = CustomBranch(custom_comp, ncomp=3)
    custom_cell = CustomCell([custom_branch], [-1])
    custom_network = CustomNetwork([custom_cell])
    custom_child_network = CustomChildNetwork([custom_cell])

    custom_child_network.cell(0).branch(0).comp(0).nodes



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jax import config

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")

import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"

from math import pi

import jax.numpy as jnp
import numpy as np
import pytest
from jax import value_and_grad

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import fully_connect
from jaxley.synapses import IonotropicSynapse, TestSynapse


@pytest.mark.slow
def test_network_grad(SimpleNet):
    net = SimpleNet(7, 5, 4)
    net.insert(HH())

    _ = np.random.seed(0)
    pre = net.cell([0, 1, 2])
    post = net.cell([3, 4, 5])
    fully_connect(pre, post, IonotropicSynapse(), random_post_comp=True)
    fully_connect(pre, post, TestSynapse(), random_post_comp=True)

    pre = net.cell([3, 4, 5])
    post = net.cell(6)
    fully_connect(pre, post, IonotropicSynapse(), random_post_comp=True)
    fully_connect(pre, post, TestSynapse(), random_post_comp=True)

    area = 2 * pi * 10.0 * 1.0
    point_process_to_dist_factor = 100_000.0 / area

    net.set("IonotropicSynapse_gS", 0.44 / point_process_to_dist_factor)
    net.set("TestSynapse_gC", 0.62 / point_process_to_dist_factor)
    net.IonotropicSynapse.edge([0, 2, 4]).set(
        "IonotropicSynapse_gS", 0.32 / point_process_to_dist_factor
    )
    net.TestSynapse.edge([0, 3, 5]).set(
        "TestSynapse_gC", 0.24 / point_process_to_dist_factor
    )

    current = jx.step_current(
        i_delay=0.5, i_dur=0.5, i_amp=0.1, delta_t=0.025, t_max=10.0
    )
    for i in range(3):
        net.cell(i).branch(0).loc(0.0).stimulate(current)

    net.cell(6).branch(0).loc(0.0).record()

    def simulate(params):
        return jnp.sum(
            jx.integrate(net, params=params, voltage_solver="jaxley.dhs.cpu")[0, ::40]
        )

    net.make_trainable("HH_gNa")
    net.cell([0, 1, 4]).make_trainable("HH_gK")
    net.cell("all").make_trainable("HH_gLeak")

    net.IonotropicSynapse.make_trainable("IonotropicSynapse_gS")
    net.TestSynapse.edge([0, 2]).make_trainable("TestSynapse_gC")

    params = net.get_parameters()
    grad_fn = value_and_grad(simulate)
    v, g = grad_fn(params)

    value_300724 = jnp.asarray(-592.26920963)
    max_error = np.max(np.abs(v - value_300724))
    tolerance = 1e-8
    assert max_error <= tolerance, f"Error is {max_error} > {tolerance}"
    grad_300724 = [
        {"HH_gNa": jnp.asarray([-1420.82493379])},
        {"HH_gK": jnp.asarray([36.76796492, 14.50837022, 127.31361882])},
        {
            "HH_gLeak": jnp.asarray(
                [
                    -1511.83061117,
                    -657.06543355,
                    -276.59064858,
                    -348.48049676,
                    -10613.6754718,
                    -1240.62600873,
                    -268475.23225046,
                ]
            )
        },
        {"IonotropicSynapse_gS": jnp.asarray([-149134.15237682])},
        {"TestSynapse_gC": jnp.asarray([-75.34438431, -163.50398586])},
    ]

    for true_g, new_g in zip(grad_300724, g):
        for key in true_g:
            max_error = np.max(np.abs(true_g[key] - new_g[key]))
            tolerance = 1e-3  # Leak cond has a huge gradient...
            assert (
                max_error <= tolerance
            ), f"Error for {key} is {max_error} > {tolerance}"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os
from copy import deepcopy

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"

import jax.numpy as jnp
import networkx as nx
import numpy as np
import pandas as pd
import pytest

import jaxley as jx
from jaxley import connect
from jaxley.channels import HH
from jaxley.channels.pospischil import K, Leak, Na
from jaxley.io.graph import (
    _add_missing_graph_attrs,
    build_compartment_graph,
    from_graph,
    to_swc_graph,
)
from jaxley.modules.base import to_graph
from jaxley.morphology import morph_connect, morph_delete
from jaxley.synapses import IonotropicSynapse, TestSynapse

# from jaxley.utils.misc_utils import recursive_compare
from tests.helpers import (
    equal_both_nan_or_empty_df,
    get_segment_xyzrL,
    import_neuron_morph,
)


# test exporting and re-importing of different modules
def test_graph_import_export_cycle(
    SimpleComp, SimpleBranch, SimpleCell, SimpleNet, SimpleMorphCell
):
    """Export and import a module and check whether .nodes and .edges remain same."""
    np.random.seed(0)
    comp = SimpleComp()
    branch = SimpleBranch(4)
    cell = SimpleCell(5, 4)
    morph_cell = SimpleMorphCell(ncomp=1)
    net = SimpleNet(3, 5, 4)

    # add synapses
    connect(net[0, 0, 0], net[1, 0, 0], IonotropicSynapse())
    connect(net[0, 0, 1], net[1, 0, 1], IonotropicSynapse())
    # connect(net[0, 0, 1], net[1, 0, 1], TestSynapse()) # makes test fail, see warning w. synapses = True

    # add groups
    net.cell(2).add_to_group("cell2")
    net.cell(2).branch(1).add_to_group("cell2branch1")

    # add ion channels
    net.cell(0).insert(Na())
    net.cell(0).insert(Leak())
    net.cell(1).branch(1).insert(Na())
    net.cell(0).insert(K())

    # test consistency of exported and re-imported modules
    for module in [comp, branch, cell, net, morph_cell]:
        module.compute_xyz()  # ensure x,y,z in nodes b4 exporting for later comparison
        module.compute_compartment_centers()

        # ensure to_graph works
        module_graph = to_graph(module, channels=True, synapses=True)

        # ensure prev exported graph can be read
        re_module = from_graph(module_graph, traverse_for_solve_order=False)

        # ensure to_graph works for re-imported modules
        re_module_graph = to_graph(re_module, channels=True, synapses=True)

        # ensure original module and re-imported module are equal
        assert np.all(equal_both_nan_or_empty_df(re_module.nodes, module.nodes))
        assert np.all(equal_both_nan_or_empty_df(re_module.edges, module.edges))
        assert np.all(
            equal_both_nan_or_empty_df(re_module.branch_edges, module.branch_edges)
        )

        for k in module.group_names:
            assert k in re_module.group_names

        for re_xyzr, xyzr in zip(re_module.xyzr, module.xyzr):
            re_xyzr[np.isnan(re_xyzr)] = -1
            xyzr[np.isnan(xyzr)] = -1

            assert np.all(re_xyzr == xyzr)

        re_imported_mechs = re_module.channels + re_module.synapses
        for re_mech, mech in zip(re_imported_mechs, module.channels + module.synapses):
            assert np.all(re_mech.name == mech.name)

        # ensure exported graph and re-exported graph are equal
        node_df = pd.DataFrame(
            [d for i, d in module_graph.nodes(data=True)],
            index=module_graph.nodes,
        )
        node_df = node_df.loc[node_df["type"] != "branchpoint"].sort_index()

        re_node_df = pd.DataFrame(
            [d for i, d in re_module_graph.nodes(data=True)],
            index=re_module_graph.nodes,
        )
        re_node_df = re_node_df.loc[re_node_df["type"] != "branchpoint"].sort_index()
        assert np.all(equal_both_nan_or_empty_df(node_df, re_node_df))

        edges = pd.DataFrame(
            [
                {
                    "pre_index": i,
                    "post_index": j,
                    **module_graph.edges[i, j],
                }
                for (i, j) in module_graph.edges
            ]
        )
        re_edges = pd.DataFrame(
            [
                {
                    "pre_index": i,
                    "post_index": j,
                    **re_module_graph.edges[i, j],
                }
                for (i, j) in re_module_graph.edges
            ]
        )
        assert np.all(equal_both_nan_or_empty_df(edges, re_edges))

        # ignore "externals", "recordings", "trainable_params", "indices_set_by_trainables"
        for k in ["ncomp"]:
            assert module_graph.graph[k] == re_module_graph.graph[k]

        # assume if module can be integrated, so can be comp, cell and branch
        if isinstance(module, jx.Network):
            # test integration of re-imported module
            re_module.select(nodes=0).record(verbose=False)
            jx.integrate(re_module, t_max=0.5)


@pytest.mark.parametrize(
    "file",
    [
        "morph_3_types_single_point_soma.swc",
        "morph_3_types.swc",
        "morph_interrupted_soma.swc",
        "morph_soma_both_ends.swc",
        "morph_somatic_branchpoint.swc",
        "morph_non_somatic_branchpoint.swc",
        "morph_ca1_n120_single_point_soma.swc",
        "morph_ca1_n120.swc",
        "morph_l5pc_with_axon.swc",
        "morph_allen_485574832.swc",
        pytest.param(
            "morph_flywire_t4_720575940626407426.swc",
            marks=pytest.mark.xfail(reason="NEURON throws .hoc error."),
        ),
        pytest.param(
            "morph_retina_20161028_1.swc",
            marks=pytest.mark.xfail(reason="Branch with l=0. Jaxley: 0.1, NEURON: 0.0"),
        ),
    ],
)
def test_trace_branches(file):
    """Test whether all branch lengths match NEURON."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)
    swc_graph = to_swc_graph(fname)

    # pre-processing
    comp_graph = build_compartment_graph(
        swc_graph, ncomp=1, ignore_swc_tracing_interruptions=False
    )

    nx_branch_lens = []
    for n in comp_graph.nodes:
        if comp_graph.nodes[n]["type"] == "comp":
            nx_branch_lens.append(comp_graph.nodes[n]["length"])
    nx_branch_lens = np.sort(nx_branch_lens)

    h, _ = import_neuron_morph(fname)
    neuron_branch_lens = np.sort([sec.L for sec in h.allsec()])

    errors = np.abs(neuron_branch_lens - nx_branch_lens)
    assert sum(errors > 1e-3) == 0


@pytest.mark.parametrize(
    "file",
    [
        "morph_ca1_n120_single_point_soma.swc",
        "morph_ca1_n120.swc",
        "morph_l5pc_with_axon.swc",
    ],
)
def test_from_graph_vs_NEURON(file):
    """Check whether comp xyzr match neuron."""
    ncomp = 8
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    swc_graph = to_swc_graph(fname)
    comp_graph = build_compartment_graph(
        swc_graph,
        ncomp=ncomp,
        max_len=2000,
        ignore_swc_tracing_interruptions=False,
    )
    cell = from_graph(comp_graph)
    cell.compute_compartment_centers()
    h, neuron_cell = import_neuron_morph(fname, ncomp=ncomp)

    # remove root branch
    jaxley_comps = cell.nodes[
        ~np.isclose(cell.nodes["length"], 0.1 / ncomp)
    ].reset_index(drop=True)

    jx_branch_lens = (
        jaxley_comps.groupby("global_branch_index")["length"].sum().to_numpy()
    )

    # match by branch lengths
    neuron_xyzd = [np.array(s.psection()["morphology"]["pts3d"]) for s in h.allsec()]
    neuron_branch_lens = np.array(
        [
            np.sqrt((np.diff(n[:, :3], axis=0) ** 2).sum(axis=1)).sum()
            for n in neuron_xyzd
        ]
    )
    neuron_inds = np.argsort(neuron_branch_lens)
    jx_inds = np.argsort(jx_branch_lens)

    neuron_df = pd.DataFrame(columns=["neuron_idx", "x", "y", "z", "radius", "length"])
    jx_df = pd.DataFrame(columns=["jx_idx", "x", "y", "z", "radius", "length"])
    for k in range(len(neuron_inds)):
        neuron_comp_k = np.array(
            [
                get_segment_xyzrL(list(h.allsec())[neuron_inds[k]], comp_idx=i)
                for i in range(ncomp)
            ]
        )
        # make this a dataframe
        neuron_comp_k = pd.DataFrame(
            neuron_comp_k, columns=["x", "y", "z", "radius", "length"]
        )
        neuron_comp_k["idx"] = neuron_inds[k]
        jx_comp_k = jaxley_comps[jaxley_comps["global_branch_index"] == jx_inds[k]][
            ["x", "y", "z", "radius", "length"]
        ]
        jx_comp_k["idx"] = jx_inds[k]
        neuron_df = pd.concat([neuron_df, neuron_comp_k], axis=0, ignore_index=True)
        jx_df = pd.concat([jx_df, jx_comp_k], axis=0, ignore_index=True)

    errors = neuron_df["neuron_idx"].to_frame()
    errors["jx_idx"] = jx_df["jx_idx"]
    errors[["x", "y", "z"]] = neuron_df[["x", "y", "z"]] - jx_df[["x", "y", "z"]]
    errors["xyz"] = np.sqrt((errors[["x", "y", "z"]] ** 2).sum(axis=1))
    errors["radius"] = neuron_df["radius"] - jx_df["radius"]
    errors["length"] = neuron_df["length"] - jx_df["length"]

    assert sum(errors.groupby("jx_idx")["xyz"].max() > 1e-3) == 0
    assert sum(errors.groupby("jx_idx")["radius"].max() > 1e-3) == 0
    assert sum(errors.groupby("jx_idx")["length"].max() > 1e-3) == 0


def test_edges_only_to_jaxley():
    """Test if edge graph can pe imported into to jaxley."""
    sets_of_edges = [
        [(0, 1), (1, 2), (2, 3)],
        [(0, 1), (1, 2), (1, 3), (2, 4), (2, 5), (3, 6), (3, 7)],
    ]
    for edges in sets_of_edges:
        graph = nx.Graph(edges)
        swc_graph = _add_missing_graph_attrs(graph)
        comp_graph = build_compartment_graph(swc_graph, ncomp=1, min_radius=1.0)
        edge_module = from_graph(comp_graph)


@pytest.mark.parametrize("ncomp", [1, 3])
def test_morph_delete(ncomp: int):
    """Test correctness of `nodes` and voltages after `morph_delete`."""
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0, 0])
    cell.branch(0).set("length", 50.0)
    cell = morph_delete(cell.branch(2))
    cell.insert(HH())

    cell2 = jx.Cell(branch, parents=[-1, 0])
    cell2.branch(0).set("length", 50.0)
    cell2.insert(HH())

    cell[0, 0].record()
    cell[0, 0].stimulate(0.1 * jnp.ones((100,)))
    cell2[0, 0].record()
    cell2[0, 0].stimulate(0.1 * jnp.ones((100,)))

    v1 = jx.integrate(cell)
    v2 = jx.integrate(cell2)
    assert np.max(np.abs(v1 - v2)) < 1e-8, "voltages do not match."

    # Drop xyz because the first cell had branches that form a "star", so even
    # after deleting a branch we do not expect xyz to be a straight line.
    assert np.all(
        equal_both_nan_or_empty_df(
            cell.nodes.drop(columns=["x", "y", "z"]),
            cell2.nodes.drop(columns=["x", "y", "z"]),
        )
    )


@pytest.mark.parametrize("ncomp", [1, 3])
def test_morph_attach(ncomp: int):
    """Test correctness of `nodes` and voltages after `morph_attach`."""
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0])
    cell.insert(Leak())
    stub = jx.Cell(branch, parents=[-1])
    stub.set("length", 80.0)
    stub.insert(HH())
    cell = morph_connect(cell.branch(1).loc(0.0), stub.branch(0).loc(0.0))

    cell2 = jx.Cell(branch, parents=[-1, 0, 0])
    cell2.branch(2).set("length", 80.0)
    cell2.branch(2).insert(HH())
    cell2.branch([0, 1]).insert(Leak())

    cell[0, 0].record()
    cell[0, 0].stimulate(0.1 * jnp.ones((100,)))
    cell2[0, 0].record()
    cell2[0, 0].stimulate(0.1 * jnp.ones((100,)))

    v1 = jx.integrate(cell)
    v2 = jx.integrate(cell2)
    assert np.max(np.abs(v1 - v2)) < 1e-8, "voltages do not match."

    # Drop xyz because the first cell had branches that form a "star", so even
    # after deleting a branch we do not expect xyz to be a straight line.
    assert np.all(
        equal_both_nan_or_empty_df(
            cell.nodes.drop(columns=["x", "y", "z"]),
            cell2.nodes.drop(columns=["x", "y", "z"]),
        )
    )


@pytest.mark.parametrize("ncomp", [1, 2])
def test_morph_edit_swc(ncomp: int):
    """Check whether we get NaN after having deleted and added things to SWC."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_l5pc_with_axon.swc")
    cell = jx.read_swc(fname, ncomp=ncomp, backend="graph")
    cell = morph_delete(cell.axon)
    cell = morph_delete(cell.apical)

    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp=ncomp)
    stub = jx.Cell(branch, parents=[-1])
    stub.set("length", 100.0)
    stub.add_to_group("stub")  # To more easily find the stub later.

    # Implicitly also tests whether it can be combined with groups (`.soma`), and
    # whether branchpoint nodes _and_ tip nodes work (branchpoint node for `cell`, tip
    # for `stub`).
    cell = morph_connect(cell.soma.branch(0).loc(1.0), stub.branch(0).loc(0.0))

    # Modify a bit and run a simulation.
    cell.stub.set_ncomp(4)
    cell.branch(3).set_ncomp(2)

    # Channels and initialization.
    cell.soma.insert(HH())
    cell.insert(Leak())
    cell.set("v", -65.0)
    cell.init_states()

    # Simulation.
    cell[0, 0].record("v")
    cell.stub.branch(0).comp(3).record("v")
    cell.soma.branch(0).comp(0).stimulate(jx.step_current(10.0, 5.0, 0.2, 0.025, 100.0))
    v = jx.integrate(cell)

    assert np.invert(np.any(np.isnan(v))), "Found NaN"


def test_trim_dendrites_of_swc():
    """This function tests whether we can successfully trim dendrites.

    It is just an API test and does not check for correctness.

    When the morphology is being trimmed, it deletes node [0] which had caused issues
    at some point.
    """

    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
    swc_graph = to_swc_graph(fname)
    comp_graph = build_compartment_graph(swc_graph, ncomp=1)

    # Next, we loop over all nodes. We want to keep nodes only if they made any of the
    # following conditions:
    # - if a node has more than one neighbor (`degree > 1`),
    # - if its compartment length is > 250 $\mu$m, or
    # - if it is a soma.
    nodes_to_keep = []
    for node in comp_graph.nodes:
        degree = comp_graph.in_degree(node) + comp_graph.out_degree(node)

        condition1 = degree > 1
        condition2 = comp_graph.nodes[node]["length"] > 250.0
        condition3 = "soma" in comp_graph.nodes[node]["groups"]
        if condition1 or condition2 or condition3:
            nodes_to_keep.append(node)

    comp_graph = nx.subgraph(comp_graph, nodes_to_keep)
    cell = from_graph(comp_graph)
    cell.delete_recordings()
    cell.delete_stimuli()
    cell.soma.branch(0).comp(0).record()
    cell.soma.branch(0).comp(0).stimulate(
        jx.step_current(10.0, 20.0, 0.1, 0.025, 100.0)
    )
    v = jx.integrate(cell)
    assert np.invert(np.any(np.isnan(v))), "Found a NaN in the voltage."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"
import numpy as np
import pytest
from neuron import h

import jaxley as jx
from jaxley.channels import HH

_ = h.load_file("stdlib.hoc")
_ = h.load_file("import3d.hoc")


@pytest.mark.parametrize(
    "file",
    [
        "morph_ca1_n120_single_point_soma.swc",
        "morph_ca1_n120.swc",
        "morph_l5pc_with_axon.swc",
        "morph_allen_485574832.swc",
        "morph_variable_radiuses_within_branch.swc",
    ],
)
def test_swc_voltages(file, SimpleMorphCell):
    """Check if voltages of SWC recording match.

    To match the branch indices between NEURON and jaxley, we rely on comparing the
    length of the branches.

    It tests whether, on average over time, the voltage is off by less than 0.5 mV
    for every recording.
    """
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "../swc_files", file)  # n120

    i_delay = 5.0
    i_dur = 20.0
    i_amp = 1.0
    t_max = 30.0
    dt = 0.025

    stim_loc = 0.51
    loc = 0.51

    if file == "morph_variable_radiuses_within_branch.swc":
        ncomp_per_branch = 1
    else:
        ncomp_per_branch = 3

    ##################### NEURON ##################
    h.secondorder = 0

    for sec in h.allsec():
        h.delete_section(sec=sec)

    cell = h.Import3d_SWC_read()
    cell.input(fname)
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)

    for sec in h.allsec():
        sec.nseg = ncomp_per_branch

    pathlengths_neuron = np.asarray([sec.L for sec in h.allsec()])

    ####################### jaxley ##################
    cell = SimpleMorphCell(
        fname,
        ncomp_per_branch,
        max_branch_len=2_000.0,
        ignore_swc_tracing_interruptions=False,
    )

    pathlengths = []
    for branch in cell.branches:
        pathlengths.append(branch.nodes["length"].sum())
    pathlengths_jaxley = np.asarray(pathlengths)

    cell.insert(HH())

    jaxley_inds = [0, 1, 2, 5, 10, 16, 20, 40, 60, 80]
    jaxley_inds = [ind for ind in jaxley_inds if ind < len(pathlengths_jaxley)]

    neuron_inds = []
    for jaxley_ind in jaxley_inds:
        for i, p in enumerate(pathlengths_jaxley):
            if i == jaxley_ind:
                closest_match = np.argmin(np.abs(pathlengths_neuron - p))
                neuron_inds.append(closest_match)

    cell.set("axial_resistivity", 100.0)
    cell.set("v", -62.0)
    cell.set("HH_m", 0.074901)
    cell.set("HH_h", 0.4889)
    cell.set("HH_n", 0.3644787)
    cell.soma.branch(0).loc(stim_loc).stimulate(
        jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    )
    for i in jaxley_inds:
        cell.branch(i).loc(loc).record(verbose=False)

    voltages_jaxley = jx.integrate(cell, delta_t=dt, voltage_solver="jaxley.dhs.cpu")

    ################### NEURON #################
    stim = h.IClamp(h.soma[0](stim_loc))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    counter = 0
    voltage_recs = {}

    for r in neuron_inds:
        for i, sec in enumerate(h.allsec()):
            if i == r:
                v = h.Vector()
                v.record(sec(loc)._ref_v)
                voltage_recs[f"v{counter}"] = v
                counter += 1

    for sec in h.allsec():
        sec.insert("hh")
        sec.Ra = 100.0

        sec.gnabar_hh = 0.120  # S/cm2
        sec.gkbar_hh = 0.036  # S/cm2
        sec.gl_hh = 0.0003  # S/cm2
        sec.ena = 50  # mV
        sec.ek = -77.0  # mV
        sec.el_hh = -54.3  # mV

    h.dt = dt
    tstop = t_max
    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < tstop:
            h.fadvance()

    initialize()
    integrate()
    voltages_neuron = np.asarray([voltage_recs[key] for key in voltage_recs])
    errors = np.mean(np.abs(voltages_jaxley - voltages_neuron), axis=1)

    ###################### check ################
    assert all(errors < 0.5), f"Error {np.max(errors)} > 0.5. Voltages do not match."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.channels.channel import Channel  # isort: skip
from jaxley.channels.hh import HH
from jaxley.channels.non_capacitive.izhikevich import Izhikevich
from jaxley.channels.non_capacitive.rate import Rate
from jaxley.channels.non_capacitive.spike import Fire
from jaxley.channels.pospischil import CaL, CaT, K, Km, Leak, Na

__all__ = [
    "Channel",
    "HH",
    "Izhikevich",
    "Rate",
    "Fire",
    "CaL",
    "CaT",
    "K",
    "Km",
    "Leak",
    "Na",
]



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os
import pickle

import pytest

import jaxley as jx
from jaxley.channels import HH
from jaxley.synapses import IonotropicSynapse

# create modules (cannot use fixtures for pickling, since they rely on local func defs)
comp = jx.Compartment()
branch = jx.Branch(comp, 4)
cell = jx.Cell([branch] * 3, [-1, 0, 0])
net = jx.Network([cell] * 2)
fname = os.path.join(os.path.dirname(__file__), "swc_files", "morph_ca1_n120.swc")
morph_cell = jx.read_swc(fname, ncomp=1, max_branch_len=2_000, assign_groups=True)

# insert mechanisms
net.cell(0).branch("all").insert(HH())
net.cell(0).branch(0).comp(0).record("v")
jx.connect(
    net.cell(0).branch(0).comp(0), net.cell(1).branch(0).comp(0), IonotropicSynapse()
)


@pytest.mark.parametrize(
    "module", [comp, branch, cell, morph_cell, net], ids=lambda x: x.__class__.__name__
)
def test_pickle(module):
    pickled = pickle.dumps(module)
    unpickled = pickle.loads(pickled)

    view = module.select(0)
    pickled = pickle.dumps(view)
    unpickled = pickle.loads(pickled)

    # assert module == unpickled # TODO: implement __eq__ for all classes



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import hashlib
import json
import os
import time
from functools import wraps

import numpy as np
import pytest
from jax import jit
from jaxley_mech.channels.l5pc import CaHVA
from scipy.stats import t as t_dist

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import sparse_connect
from jaxley.pumps import CaFaradayConcentrationChange, CaNernstReversal
from jaxley.synapses import IonotropicSynapse

pytestmark = pytest.mark.regression  # mark all tests as regression tests in this file

# Every runtime test needs to have the following structure:
#
# @compare_to_baseline()
# def test_runtime_of_x(**kwargs) -> Dict:
#   t1 = time.time()
#   time.sleep(0.1)
#   # do something
#   t2 = time.time()
#   # do something else
#   t3 = time.time()
#   return {"sth": t2-t1, sth_else: t3-t2}

# The test should return a dictionary with the runtime of each part of the test.
# This way the runtimes of different parts of the code can be compared to each other.

# The @compare_to_baseline decorator will compare the runtime of the test to the baseline
# and raise an assertion error if the runtime is significantly slower than the baseline.
# The decorator will also save the runtime of the test to a database, which will be used
# to generate a report comparing the runtime of the tests to the baseline. The database
# takes into account the input_kwargs of the test, the name of the test and the runtimes
# of each part.

NEW_BASELINE = int(os.environ["NEW_BASELINE"]) if "NEW_BASELINE" in os.environ else 0
CONFIDENCE = 0.95

dirname = os.path.dirname(__file__)
fpath_baselines = os.path.join(dirname, "regression_test_baselines.json")
fpath_results = os.path.join(dirname, "regression_test_results.json")
with open(fpath_results, "w") as f:  # clear previous results
    f.write("{}")


def generate_regression_report(base_results, new_results):
    """Compare two sets of benchmark results and generate a diff report."""
    report = []
    for key in new_results:
        new_data = new_results[key]
        base_data = base_results.get(key)
        kwargs = ", ".join([f"{k}={v}" for k, v in new_data["input_kwargs"].items()])
        func_name = new_data["test_name"]
        func_signature = f"{func_name}({kwargs})"

        new_runtimes = new_data["runtimes"]
        base_runtimes = (
            {k: None for k in new_data.keys()}
            if base_data is None
            else base_data["runtimes"]
        )

        report.append(func_signature)
        for key, new_time in new_runtimes.items():
            base_time = base_runtimes.get(key)
            diff = None if base_time is None else ((new_time - base_time) / base_time)

            status = ""
            if base_time is None:
                status = "🆕"
            elif new_time <= base_time:
                status = "🟢" if diff is not None and diff < 0.05 else "🟠"
            elif new_time > base_time:
                status = "🔴"
            else:
                status = "❌"  # This should never happen.

            time_str = (
                f"({new_time:.3f}s)"
                if diff is None
                else f"({diff:+.2%} vs {base_time:.3f}s)"
            )
            report.append(f"{status} {key}: {time_str}.")
        report.append("")

    return "\n".join(report)


def generate_unique_key(d):
    # Generate a unique key for each test case. Makes it possible to compare tests
    # with different input_kwargs.
    hash_obj = hashlib.sha256(bytes(json.dumps(d, sort_keys=True), encoding="utf-8"))
    hash = hash_obj.hexdigest()
    return str(hash)


def compute_conf_bounds(X):
    df = len(X) - 1  # degrees of freedom = n-1
    critical_value = t_dist.ppf(CONFIDENCE, df)
    return np.mean(X) + critical_value * np.std(X, ddof=1)  # sample std


def load_json(fpath):
    dct = {}
    if os.path.exists(fpath):
        with open(fpath, "r") as f:
            dct = json.load(f)
    return dct


def append_to_json(fpath, test_name, input_kwargs, runtimes):
    header = {"test_name": test_name, "input_kwargs": input_kwargs}
    data = {generate_unique_key(header): {**header, "runtimes": runtimes}}

    # Save data to a JSON file
    result_data = load_json(fpath)
    result_data.update(data)

    with open(fpath, "w") as f:
        json.dump(result_data, f, indent=2)


class compare_to_baseline:
    def __init__(self, baseline_iters=5, test_iters=1):
        self.baseline_iters = baseline_iters
        self.test_iters = test_iters
        self.baselines = load_json(fpath_baselines)

    def __call__(self, func):
        @wraps(func)  # ensures kwargs exposed to pytest
        def test_wrapper(**kwargs):
            header = {"test_name": func.__name__, "input_kwargs": kwargs}
            key = generate_unique_key(header)

            runs = []
            num_iters = self.baseline_iters if NEW_BASELINE else self.test_iters
            for _ in range(num_iters):
                runtimes = func(**kwargs)
                runs.append(runtimes)

            # the baseline time is taken as the upper bound of the confidence interval,
            # while the runtimes that we test against the baseline are taken as the mean
            agg = compute_conf_bounds if NEW_BASELINE else np.mean
            runtimes = {k: agg([d[k] for d in runs]) for k in runs[0]}

            append_to_json(
                fpath_results, header["test_name"], header["input_kwargs"], runtimes
            )

            if not NEW_BASELINE:
                assert key in self.baselines, f"No baseline found for {header}"
                func_baselines = self.baselines[key]["runtimes"]
                for key, baseline in func_baselines.items():
                    assert (
                        runtimes[key] <= baseline
                    ), f"{key} is significantly slower than the baseline at {runtimes[key]:.3f}s vs. {baseline:.3f}s."

        return test_wrapper


def build_net(
    num_cells,
    artificial=True,
    connect=True,
    connection_prob=0.0,
    calcium_dynamics=False,
    voltage_solver="jaxley.dhs",
):
    _ = np.random.seed(1)  # For sparse connectivity matrix.

    if artificial:
        comp = jx.Compartment()
        branch = jx.Branch(comp, 2)
        depth = 3
        parents = [-1] + [b // 2 for b in range(0, 2**depth - 2)]
        cell = jx.Cell(branch, parents=parents)
    else:
        dirname = os.path.dirname(__file__)
        fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
        cell = jx.read_swc(fname, ncomp=4)

    if voltage_solver == "jaxley.dhs.gpu":
        cell._init_solver_jaxley_dhs_solve(allowed_nodes_per_level=16)

    if voltage_solver == "jaxley.dhs.gpu":
        net = jx.Network([cell for _ in range(num_cells)], vectorize_cells=True)
    else:
        net = jx.Network([cell for _ in range(num_cells)], vectorize_cells=False)
    # Channels.
    net.insert(HH())

    if calcium_dynamics:
        net.insert(CaHVA())
        net.insert(CaFaradayConcentrationChange())
        net.insert(CaNernstReversal())
        net.diffuse("CaCon_i")

    # Synapses.
    if connect:
        sparse_connect(
            net.cell("all"), net.cell("all"), IonotropicSynapse(), connection_prob
        )

    # Recordings.
    net[0, 1, 0].record(verbose=False)

    # Trainables.
    net.make_trainable("radius", verbose=False)
    params = net.get_parameters()

    net.to_jax()
    return net, params


@pytest.mark.parametrize(
    "num_cells, artificial, connect, connection_prob, voltage_solver, calcium_dynamics",
    (
        # Test a single SWC cell with both solvers.
        pytest.param(1, False, False, 0.0, "jaxley.dhs", False),
        pytest.param(1, False, False, 0.0, "jax.sparse", False),
        # Test a single SWC cell, and add detailed calcium mechanisms.
        pytest.param(1, False, False, 0.0, "jaxley.dhs", True),
        # Test a network of SWC cells with both solvers.
        pytest.param(10, False, True, 0.1, "jaxley.dhs", False),
        pytest.param(10, False, True, 0.1, "jax.sparse", False),
        # Test a larger network of smaller neurons with both solvers.
        pytest.param(1000, True, True, 0.001, "jaxley.dhs", False),
        pytest.param(1000, True, True, 0.001, "jaxley.dhs.gpu", False),
        pytest.param(1000, True, True, 0.001, "jax.sparse", False),
    ),
)
@compare_to_baseline(baseline_iters=3)
def test_runtime(
    num_cells: int,
    artificial: bool,
    connect: bool,
    connection_prob: float,
    voltage_solver: str,
    calcium_dynamics: bool,
):
    delta_t = 0.025
    t_max = 100.0

    def simulate(params):
        return jx.integrate(
            net,
            params=params,
            t_max=t_max,
            delta_t=delta_t,
            voltage_solver=voltage_solver,
        )

    runtimes = {}

    start_time = time.time()
    net, params = build_net(
        num_cells,
        artificial=artificial,
        connect=connect,
        connection_prob=connection_prob,
        calcium_dynamics=calcium_dynamics,
        voltage_solver=voltage_solver,
    )
    runtimes["build_time"] = time.time() - start_time

    jitted_simulate = jit(simulate)

    start_time = time.time()
    _ = jitted_simulate(params).block_until_ready()
    runtimes["compile_time"] = time.time() - start_time
    params[0]["radius"] = params[0]["radius"].at[0].set(0.5)

    start_time = time.time()
    _ = jitted_simulate(params).block_until_ready()
    runtimes["run_time"] = time.time() - start_time
    return runtimes  # @compare_to_baseline decorator will compare this to the baseline



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax.lax import ScatterDimensionNumbers, scatter_add
from jax.typing import ArrayLike


def gather_synapes(
    number_of_compartments: ArrayLike,
    post_syn_comp_inds: np.ndarray,
    current_each_synapse_voltage_term: ArrayLike,
    current_each_synapse_constant_term: ArrayLike,
) -> tuple[Array, Array]:
    """Compute current at the post synapse.

    All this does it that it sums the synaptic currents that come into a particular
    compartment. It returns an array of as many elements as there are compartments.
    """
    incoming_currents_voltages = jnp.zeros((number_of_compartments,))
    incoming_currents_content = jnp.zeros((number_of_compartments,))

    dnums = ScatterDimensionNumbers(
        update_window_dims=(),
        inserted_window_dims=(0,),
        scatter_dims_to_operand_dims=(0,),
    )
    incoming_currents_voltages = scatter_add(
        incoming_currents_voltages,
        post_syn_comp_inds[:, None],
        current_each_synapse_voltage_term,
        dnums,
    )
    incoming_currents_content = scatter_add(
        incoming_currents_content,
        post_syn_comp_inds[:, None],
        current_each_synapse_constant_term,
        dnums,
    )
    return incoming_currents_voltages, incoming_currents_content



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect
from jaxley.solver_gate import exponential_euler
from jaxley.synapses import IonotropicSynapse


@pytest.mark.parametrize("x_inf", [3.0, 30.0])
def test_exp_euler(x_inf):
    """Test whether `dx = -(x - x_inf) / x_tau` is solved correctly."""
    x = 20.0
    x_tau = 12.0
    dt = 0.1
    vectorfield = -(x - x_inf) / x_tau
    fwd_euler = x + dt * vectorfield
    exp_euler = exponential_euler(x, dt, x_inf, x_tau)
    assert np.abs(fwd_euler - exp_euler) / np.abs(fwd_euler) < 1e-4


def test_fwd_euler_and_crank_nicolson(SimpleNet):
    """Compare accuracies of different solvers."""
    net = SimpleNet(2, 3, 4, connect=True)

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    net.cell(0).branch(0).comp(0).stimulate(current)
    net.cell(1).branch(0).comp(3).record()

    net.IonotropicSynapse.set("IonotropicSynapse_gS", 0.01)
    net.insert(HH())

    # As expected, using significantly shorter compartments or lower r_a leads to NaN
    # in forward Euler.
    net.set("length", 50.0)
    net.set("axial_resistivity", 200.0)

    bwd_v = jx.integrate(net, solver="bwd_euler")
    fwd_v = jx.integrate(net, solver="fwd_euler")
    crank_v = jx.integrate(net, solver="crank_nicolson")
    exp_v = jx.integrate(net, solver="exp_euler")

    # These allowed voltage differences may appear large, but due to the steepness
    # of a spike these values actually correspond to sub-millisecond spike time
    # differences.
    assert np.max(np.abs(bwd_v - fwd_v)) < 12.0
    assert np.max(np.abs(bwd_v - crank_v)) < 6.0
    assert np.max(np.abs(bwd_v - exp_v)) < 4.0


def test_exp_euler_solver_customization(SimpleCell):
    cell = SimpleCell(2, 3)
    delta_t = 0.025

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=delta_t, t_max=5.0
    )
    cell.branch(0).comp(0).stimulate(current)
    cell.branch(0).comp(1).record()
    cell.insert(HH())

    cell.customize_solver_exp_euler(
        exp_euler_transition=cell.build_exp_euler_transition_matrix(delta_t)
    )
    v = jx.integrate(cell, solver="exp_euler")
    assert np.invert(np.any(np.isnan(v)))



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, Optional

import jax.numpy as jnp

from jaxley.synapses.synapse import Synapse


class TanhConductanceSynapse(Synapse):
    """
    Compute synaptic current for a simple conductance synapse (no state).

    This is just the TanhRateSynapse with current affected by the postsynaptic
    conductance.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        prefix = self._name
        self.synapse_params = {
            f"{prefix}_gS": 1e-4,
            f"{prefix}_e_syn": 0.0,  # mV
            f"{prefix}_x_offset": -70.0,
            f"{prefix}_slope": 1.0,
        }
        self.synapse_states = {}

    def update_states(
        self,
        states: Dict,
        delta_t: float,
        pre_voltage: float,
        post_voltage: float,
        params: Dict,
    ) -> Dict:
        """Return updated synapse state and current."""
        return {}

    def compute_current(
        self, states: Dict, pre_voltage: float, post_voltage: float, params: Dict
    ) -> float:
        """Return updated synapse state and current."""
        prefix = self._name
        tanh_pre_voltage = jnp.tanh(
            (pre_voltage - params[f"{prefix}_x_offset"]) * params[f"{prefix}_slope"]
        )
        current = (
            tanh_pre_voltage
            * params[f"{prefix}_gS"]
            * (post_voltage - params[f"{prefix}_e_syn"])
        )
        return current



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Callable, Dict, List, Optional, Tuple

import jax.numpy as jnp
import numpy as np
import pandas as pd
from matplotlib.axes import Axes

from jaxley.modules.base import Module
from jaxley.utils.cell_utils import compute_children_and_parents
from jaxley.utils.misc_utils import cumsum_leading_zero
from jaxley.utils.morph_attributes import (
    compute_axial_conductances,
    cylinder_area,
    cylinder_resistive_load,
    cylinder_volume,
    morph_attrs_from_xyzr,
    split_xyzr_into_equal_length_segments,
)


class Compartment(Module):
    """A single compartment.

    This class defines a single compartment that can be simulated by itself or
    connected up into branches. It is the basic building block of a neuron model.
    """

    compartment_params: Dict = {
        "length": 10.0,  # um
        "radius": 1.0,  # um
        "axial_resistivity": 5_000.0,  # ohm cm
        "capacitance": 1.0,  # uF/cm^2
    }
    compartment_states: Dict = {"v": -70.0}

    def __init__(self):
        """Initialize a compartment."""
        super().__init__()

        self.ncomp = 1
        self.ncomp_per_branch = np.asarray([1])
        self.total_nbranches = 1
        self.nbranches_per_cell = [1]
        self._cumsum_nbranches = np.asarray([0, 1])
        self.cumsum_ncomp = cumsum_leading_zero(self.ncomp_per_branch)
        self._n_nodes = 1

        # Setting up the `nodes` for indexing.
        self.nodes = pd.DataFrame(
            dict(global_cell_index=[0], global_branch_index=[0], global_comp_index=[0])
        )
        self._append_params_and_states(self.compartment_params, self.compartment_states)

        # Synapses.
        self.branch_edges = pd.DataFrame(
            dict(parent_branch_index=[], child_branch_index=[])
        )

        # For morphology indexing.
        self._par_inds, self._child_inds, self._child_belongs_to_branchpoint = (
            compute_children_and_parents(self.branch_edges)
        )
        self._internal_node_inds = jnp.asarray([0])

        # Coordinates.
        self.xyzr = [float("NaN") * np.zeros((2, 4))]

        l = self.compartment_params["length"]
        r = self.compartment_params["radius"]
        # l/2 because we want the input load (left half of the cylinder) and
        # the output load (right half of the cylinder).
        resistive_load = cylinder_resistive_load(l / 2, r)
        self.nodes["area"] = cylinder_area(l, r)
        self.nodes["volume"] = cylinder_volume(l, r)
        self.nodes["resistive_load_in"] = resistive_load
        self.nodes["resistive_load_out"] = resistive_load

        # Initialize the module.
        self.initialize()

    def _init_comp_graph(self):
        """Initialize attributes concerning the compartment graph.

        In particular, it initializes:
        - `_comp_edges`
        - `_branchpoints`
        - `_n_nodes`
        - `_off_diagonal_inds`

        It also initializes `_comp_edges_in_view` and `_branchpoints_in_view`."""
        # Compartment edges.
        self._comp_edges = pd.DataFrame().from_dict(
            {"source": [], "sink": [], "ordered": [], "type": []}
        )

        # Branchpoints.
        self._branchpoints = pd.DataFrame().from_dict({"x": [], "y": [], "z": []})

        # Mapping from global_comp_index to `nodes.index`.
        self._n_nodes = 1
        self._off_diagonal_inds = np.asarray([])



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from typing import List, Optional

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import Array
from jax.typing import ArrayLike


def morph_attrs_from_xyzr(
    xyzr: np.ndarray,
    min_radius: Optional[float],
    ncomp: int,
) -> float:
    """Return radius, area, volume, and resistive loads of a comp given its SWC xyzr.

    Args:
        radius_fns: Functions which, given compartment locations return the radius.
        branch_indices: The indices of the branches for which to return the radiuses.
        min_radius: If passed, the radiuses are clipped to be at least as large.
        ncomp: The number of compartments that every branch is discretized into.
    """
    # Extract 3D coordinates and radii
    positions = xyzr[:, :3]  # shape (N, 3): x, y, z
    radii = xyzr[:, 3]  # shape (N,): radius at each point

    if len(xyzr) > 1:
        # Compute Euclidean distances between consecutive points
        position_deltas = np.diff(positions, axis=0)  # shape (N-1, 3)
        segment_lengths = np.linalg.norm(position_deltas, axis=1)  # shape (N-1,)

        avg_radius = swc_radius(segment_lengths, radii)
        total_surface_area = swc_area(segment_lengths, radii)
        total_volume = swc_volume(segment_lengths, radii)

        # Finally, we compute the input and output resistive loads. For this, we first
        # have to split the xyzr into two: the ones on the left half of the
        # compartment, and the ones on the right half.
        xyzr_split = split_xyzr_into_equal_length_segments(xyzr, 2)
        resistive_load = []
        for xyzr_half in xyzr_split:
            # Extract 3D coordinates and radii
            positions_half = xyzr_half[:, :3]  # shape (N, 3): x, y, z
            radii_half = xyzr_half[:, 3]  # shape (N,): radius at each point

            # Compute Euclidean distances between consecutive points
            position_deltas_half = np.diff(positions_half, axis=0)  # shape (N-1, 3)
            segment_lengths_half = np.linalg.norm(
                position_deltas_half, axis=1
            )  # shape (N-1,)
            resistive_load.append(swc_resistive_load(segment_lengths_half, radii_half))
    else:
        avg_radius = radii.mean()
        total_surface_area = 4 * np.pi * radii[0] ** 2 / ncomp  # Surface of a sphere.
        total_volume = 4 / 3 * np.pi * radii[0] ** 3 / ncomp  # Volume of a sphere.

        # Resistive load.
        # For single point, the total length of a branch is: length = radius.
        # Thus, the length of a compartment is `radius/ncomp`.
        length = radii[0] / ncomp
        resistive_load = [length / radii[0] ** 2 / np.pi] * 2

    if min_radius is None:
        assert (
            avg_radius > 0.0
        ), "Radius 0.0 in SWC file. Set `read_swc(..., min_radius=...)`."
    else:
        avg_radius = (
            min_radius
            if (avg_radius < min_radius or np.isnan(avg_radius))
            else avg_radius
        )
    return avg_radius, total_surface_area, total_volume, *resistive_load


def split_xyzr_into_equal_length_segments(
    xyzr: np.ndarray, ncomp: int
) -> List[np.ndarray]:
    """Split xyzr into equal-length segments by inserting interpolated points as needed.

    This function was written by ChatGPT, based on the prompt:
    ```I have an array of shape 100x3. The 3 indicate x, y, z coordinates. I want to
    split this array into 4 segments, each with equal euclidean length. To have
    euclidean length exactly equal, I would like to insert additional points into
    the 100x3 array (to make it length 100 + 4 segments - 1). These points should be
    linear interpolation of neighboring points. In the final split array, the newly
    inserted nodes should be the last point of one segment and the first point of
    another segment.```

    Args:
        points: Array of 3D coordinates representing a path.
        num_segments: Number of segments to split the path into.

    Returns:
        A list of `num_segments` arrays, each containing the 3D coordinates
        of one segment. The segments have (approximately) equal Euclidean
        length, and split points are interpolated between original points.
    """
    if len(xyzr) == 1:
        return [xyzr] * ncomp

    # Compute distances between consecutive points
    xyz = xyzr[:, :3]

    # Compute distances and cumulative distances
    deltas = np.diff(xyz, axis=0)
    dists = np.linalg.norm(deltas, axis=1)
    cum_dists = np.concatenate([[0], np.cumsum(dists)])
    total_length = cum_dists[-1]

    # Target cumulative distances where we want to split
    target_dists = np.linspace(0, total_length, ncomp + 1)

    # Find insertion indices and interpolation factors
    idxs = np.searchsorted(cum_dists, target_dists, side="right") - 1
    idxs = np.clip(idxs, 0, len(xyz) - 2)  # Ensure valid indices
    local_dist = target_dists - cum_dists[idxs]

    # When two traced SWC are right on top of each other, their dists=0. Then, when
    # these points lie exactly at the point where the xyz is supposed to be split,
    # we get `segment_lens=0`, which causes frac to be infinity.
    dists = np.where(dists < 1e-14, 1e-14, dists)
    segment_lens = dists[idxs]
    frac = (local_dist / segment_lens)[:, None]  # shape (n, 1)

    # Interpolate split points
    split_points = xyzr[idxs] + frac * (xyzr[idxs + 1] - xyzr[idxs])

    # Build final list of points with inserted nodes
    all_points = [split_points[0]]
    compartment_xyzrs = []

    for i in range(1, len(split_points)):
        # Collect original points between splits.
        mask = (cum_dists > target_dists[i - 1]) & (cum_dists < target_dists[i])
        between_points = xyzr[mask]
        segment = np.vstack([all_points[-1], *between_points, split_points[i]])
        compartment_xyzrs.append(segment)
        all_points.append(split_points[i])
    return compartment_xyzrs


def swc_radius(lengths: np.ndarray, radii: np.ndarray) -> np.ndarray:
    r"""Return the radius of a branch given its SWC coordinates.

    This function computes the average radius, weighted by the length between
    two SWC points.

    Args:
        lengths: Array of shape `(N-1)`, indicating the spacing between all SWC
            points within the branch.
        radii: Array of shape `(N)`, indicating the radius of each SWC point.

    Returns:
        A radius as a scalar value."""
    radius_weights = np.zeros(len(lengths) + 1)
    radius_weights[1:] += lengths
    radius_weights[:-1] += lengths
    radius_weights /= np.sum(radius_weights)
    return np.sum(radii * radius_weights)


def swc_area(lengths: np.ndarray, radii: np.ndarray) -> np.ndarray:
    r"""Return the surface area of a compartment given its SWC coordinates.

    This function makes a truncated cone approximation between any two SWC points
    and then computes the surface area.

    Args:
        lengths: Array of shape `(N-1)`, indicating the spacing between all SWC
            points within the compartment.
        radii: Array of shape `(N)`, indicating the radius of each SWC point.

    Returns:
        A membrane surface area as a scalar value."""
    radius_start = radii[:-1]
    radius_end = radii[1:]
    delta_radii = radius_end - radius_start
    slant_lengths = np.sqrt(delta_radii**2 + lengths**2)
    frustum_surface_areas = np.pi * (radius_start + radius_end) * slant_lengths
    return np.sum(frustum_surface_areas)


def swc_volume(lengths: np.ndarray, radii: np.ndarray) -> np.ndarray:
    r"""Return the volume of a compartment given its SWC coordinates.

    This function makes a truncated cone approximation between any two SWC points
    and then computes the volume. This function is used only for ion diffusion.

    Args:
        lengths: Array of shape `(N-1)`, indicating the spacing between all SWC
            points within the compartment.
        radii: Array of shape `(N)`, indicating the radius of each SWC point.

    Returns:
        A volume as a scalar value."""
    radius_start = radii[:-1]
    radius_end = radii[1:]
    volume = (
        (np.pi / 3)
        * lengths
        * (radius_start**2 + radius_start * radius_end + radius_end**2)
    )
    return np.sum(volume)


def swc_resistive_load(lengths: np.ndarray, radii: np.ndarray) -> np.ndarray:
    r"""Return the resistive load of a compartment given its SWC coordinates.

    The resistive load is defined as the integral over :math:`1/(\pi r^2)`, i.e.,

    .. math::

        r_l = \frac{1}{\pi} \int \frac{1}{r^2} \, dl

    As an example, if the radius is constant, then we obtain :math:`l / r^2 / \pi`,
    which corresponds exactly to the length divided by the cross section.

    This function makes a truncated cone approximation between any two SWC points
    and then computes the resistive load with the equation above. In the function
    `compute_axial_conductances()`, the resistive load gets multiplied by the
    axial resistivity (r_a, in :math:`\ohm` cm) to obtain the resistance between
    compartments.

    Args:
        lengths: Array of shape `(N-1)`, indicating the spacing between all SWC
            points within the compartment.
        radii: Array of shape `(N)`, indicating the radius of each SWC point.

    Returns:
        A resistive load as a scalar value."""
    lengths = np.asarray(lengths)
    radius_start = np.asarray(radii[:-1])
    radius_end = np.asarray(radii[1:])
    delta_radius = radius_end - radius_start

    segment_integrals = np.empty_like(lengths)

    # Segments with constant radius.
    is_constant_radius = np.isclose(delta_radius, 0)
    segment_integrals[is_constant_radius] = (
        lengths[is_constant_radius] / radius_start[is_constant_radius] ** 2
    )

    # Segments with varying radius (truncated cone).
    is_varying_radius = ~is_constant_radius
    segment_integrals[is_varying_radius] = (
        lengths[is_varying_radius]
        / delta_radius[is_varying_radius]
        * (1 / radius_start[is_varying_radius] - 1 / radius_end[is_varying_radius])
    )

    return np.sum(segment_integrals) / jnp.pi


def cylinder_area(length: ArrayLike, radius: ArrayLike) -> Array:
    r"""Return the surface area of a cylindric compartment, given its length and radius.

    Args:
        lengths: The lengths of M cylindric compartments, shape (M,).
        radii: The radii of M cylindric compartments, shape (M,).

    Returns:
        The membrane surface area of each M cylindric compartments, shape (M,)."""
    return 2.0 * jnp.pi * radius * length


def cylinder_volume(length: ArrayLike, radius: ArrayLike) -> Array:
    r"""Return the volume of a cylindric compartment, given its length and radius.

    The resistive load is defined as the integral over :math:`1/(\pi r^2)`, i.e.,

    .. math::

        r_l = \frac{1}{\pi} \int \frac{1}{r^2} \, dl

    For a cylinder, the radius is constant, so we obtain :math:`l / r^2 / \pi`.
    This corresponds exactly to the length divided by the cross section.

    Args:
        lengths: The lengths of M cylindric compartments, shape (M,).
        radii: The radii of M cylindric compartments, shape (M,).

    Returns:
        The volume of each M cylindric compartments, shape (M,)."""
    return length * radius**2 * jnp.pi


def cylinder_resistive_load(length: ArrayLike, radius: ArrayLike) -> Array:
    r"""Return the resistive load of a cylindric compartment, given length and radius.

    The resistive load is defined as the integral over :math:`1/(\pi r^2)`, i.e.,

    .. math::

        r_l = \frac{1}{\pi} \int \frac{1}{r^2} \, dl

    For a cylinder, the radius is constant, so we obtain :math:`l / r^2 / \pi`.
    This corresponds exactly to the length divided by the cross section.

    Args:
        lengths: The lengths of M cylindric compartments, shape (M,).
        radii: The radii of M cylindric compartments, shape (M,).

    Returns:
        The resistive load of each M cylindric compartments, shape (M,)."""
    return length / radius**2 / jnp.pi


def compute_axial_conductances(
    comp_edges: pd.DataFrame,
    params: dict[str, Array],
    diffusion_states: list[str],
) -> dict[str, Array]:
    r"""Given `comp_edges`, radius, length, r_a, cm, compute the axial conductances.

    Note that the resulting axial conductances will already by divided by the
    capacitance `cm`.
    """
    ordered_conds = jnp.zeros((1 + len(diffusion_states), len(comp_edges)))

    axial_conds = jnp.stack(
        [1 / params["axial_resistivity"]]
        + [params[f"axial_diffusion_{d}"] for d in diffusion_states]
    )
    # These are still _compartment_ properties.
    comp_source_r_a = params["resistive_load_out"] / axial_conds
    comp_sink_r_a = params["resistive_load_in"] / axial_conds

    # comp_r_a has shape (N, 2, num_comps). Here, N is the number of states that are
    # diffused (including voltage).
    comp_r_a = jnp.stack([comp_sink_r_a, comp_source_r_a], axis=1)

    # `Compartment-to-compartment` (c2c) axial coupling conductances.
    condition = comp_edges["type"].to_numpy() == 0
    source_comp_inds = np.asarray(comp_edges[condition]["source"].to_list()).astype(int)
    sink_comp_inds = np.asarray(comp_edges[condition]["sink"].to_list()).astype(int)
    ordered_edge = np.asarray(comp_edges[condition]["ordered"].to_list())

    # Now we compute c2c _comp_edges_ properties.
    if len(sink_comp_inds) > 0:
        r_a_of_sources = comp_r_a[:, ordered_edge, source_comp_inds]
        r_a_of_sinks = comp_r_a[:, 1 - ordered_edge, sink_comp_inds]
        r_a = r_a_of_sources + r_a_of_sinks

        # Voltage diffusion.
        conds_c2c = 1 / r_a[:1] / params["area"][sink_comp_inds]

        # We only divide the axial _voltage_ conductances by the
        # capacitance, _not_ the axial conductances of the diffusing ions.
        conds_c2c /= params["capacitance"][sink_comp_inds]
        # Multiply by 10**7 to convert (S / cm / um) -> (mS / cm^2).
        conds_c2c *= 10**7

        # For ion diffusion, we have to divide by the volume, not the surface area.
        conds_diffusion = 1 / r_a[1:] / params["volume"][sink_comp_inds]
        conds_c2c = jnp.concatenate([conds_c2c, conds_diffusion], axis=0)

        inds = jnp.asarray(comp_edges[condition].index)
        ordered_conds = ordered_conds.at[:, inds].set(conds_c2c)

    # `branchpoint-to-compartment` (bp2c) axial coupling conductances.
    condition = comp_edges["type"].isin([1, 2])
    sink_comp_inds = np.asarray(comp_edges[condition]["sink"].to_list()).astype(int)
    ordered_edge = np.asarray(comp_edges[condition]["ordered"].to_list())

    if len(sink_comp_inds) > 0:
        r_a = comp_r_a[:, 1 - ordered_edge, sink_comp_inds]

        # Voltage diffusion.
        conds_bp2c = 1 / r_a[:1] / params["area"][sink_comp_inds]
        conds_bp2c /= params["capacitance"][sink_comp_inds]
        # Multiply by 10**7 to convert (S / cm / um) -> (mS / cm^2).
        conds_bp2c *= 10**7

        # For ion diffusion, we have to divide by the volume, not the surface area.
        conds_diffusion = 1 / r_a[1:] / params["volume"][sink_comp_inds]
        conds_bp2c = jnp.concatenate([conds_bp2c, conds_diffusion], axis=0)

        inds = jnp.asarray(comp_edges[condition].index)
        ordered_conds = ordered_conds.at[:, inds].set(conds_bp2c)

    # `compartment-to-branchpoint` (c2bp) axial coupling conductances.
    condition = comp_edges["type"].isin([3, 4])
    source_comp_inds = np.asarray(comp_edges[condition]["source"].to_list()).astype(int)

    comp_source_g_a = 1 / params["resistive_load_out"] * axial_conds
    comp_sink_g_a = 1 / params["resistive_load_in"] * axial_conds
    comp_g_a = jnp.stack([comp_sink_g_a, comp_source_g_a], axis=1)

    if len(source_comp_inds) > 0:
        conds_c2bp = comp_g_a[:, 1 - ordered_edge, source_comp_inds]
        inds = jnp.asarray(comp_edges[condition].index)
        ordered_conds = ordered_conds.at[:, inds].set(conds_c2bp)

    # Reformat the conductances along the key of the quantity being diffused.
    ordered_conds_as_dict = {}
    for i, key in enumerate(["v"] + diffusion_states):
        ordered_conds_as_dict[key] = ordered_conds[i]

    return ordered_conds_as_dict



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import connect
from jaxley.integrate import build_init_and_step_fn
from jaxley.synapses import IonotropicSynapse


def test_api_equivalence_morphology(SimpleComp):
    """Test the API for how one can build morphologies from scratch."""
    ncomp_per_branch = 2
    depth = 2
    dt = 0.025

    parents = [-1] + [b // 2 for b in range(0, 2**depth - 2)]
    parents = jnp.asarray(parents)
    num_branches = len(parents)

    comp = SimpleComp()

    branch1 = jx.Branch([comp for _ in range(ncomp_per_branch)])
    cell1 = jx.Cell([branch1 for _ in range(num_branches)], parents=parents)

    branch2 = jx.Branch(comp, ncomp=ncomp_per_branch)
    cell2 = jx.Cell(branch2, parents=parents)

    cell1.branch(2).loc(0.4).record()
    cell2.branch(2).loc(0.4).record()

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    cell1.branch(1).loc(1.0).stimulate(current)
    cell2.branch(1).loc(1.0).stimulate(current)

    voltages1 = jx.integrate(cell1, delta_t=dt)
    voltages2 = jx.integrate(cell2, delta_t=dt)
    assert (
        jnp.max(jnp.abs(voltages1 - voltages2)) < 1e-8
    ), "Voltages do not match between morphology APIs."


def test_solver_backends_comp(SimpleComp):
    """Test whether ways of adding synapses are equivalent."""
    comp = SimpleComp()

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    comp.stimulate(current)
    comp.record()

    voltages_jx_cpu = jx.integrate(comp, voltage_solver="jaxley.dhs.cpu")
    voltages_jx_gpu = jx.integrate(comp, voltage_solver="jaxley.dhs.gpu")
    voltages_jx_sparse = jx.integrate(comp, voltage_solver="jax.sparse")

    message = "Voltages do not match between"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_gpu))
    assert max_error < 1e-8, f"{message} cpu/gpu. Error={max_error}"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_sparse))
    assert max_error < 1e-8, f"{message} cpu/sparse. Error={max_error}"


def test_solver_backends_branch(SimpleBranch):
    """Test whether ways of adding synapses are equivalent."""
    branch = SimpleBranch(4)

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    branch.loc(0.0).stimulate(current)
    branch.loc(0.5).record()

    voltages_jx_cpu = jx.integrate(branch, voltage_solver="jaxley.dhs.cpu")
    voltages_jx_gpu = jx.integrate(branch, voltage_solver="jaxley.dhs.gpu")
    voltages_jx_sparse = jx.integrate(branch, voltage_solver="jax.sparse")

    message = "Voltages do not match between"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_gpu))
    assert max_error < 1e-8, f"{message} cpu/gpu. Error={max_error}"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_sparse))
    assert max_error < 1e-8, f"{message} cpu/sparse. Error={max_error}"


@pytest.mark.slow
def test_solver_backends_cell(SimpleCell):
    """Test whether ways of adding synapses are equivalent."""
    cell = SimpleCell(4, 4)

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    cell.branch(0).loc(0.0).stimulate(current)
    cell.branch(0).loc(0.5).record()
    cell.branch(3).loc(0.5).record()

    voltages_jx_cpu = jx.integrate(cell, voltage_solver="jaxley.dhs.cpu")
    voltages_jx_gpu = jx.integrate(cell, voltage_solver="jaxley.dhs.gpu")
    voltages_jx_sparse = jx.integrate(cell, voltage_solver="jax.sparse")

    message = "Voltages do not match between"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_gpu))
    assert max_error < 1e-8, f"{message} cpu/gpu. Error={max_error}"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_sparse))
    assert max_error < 1e-8, f"{message} cpu/sparse. Error={max_error}"


def test_solver_backends_net(SimpleNet):
    """Test whether ways of adding synapses are equivalent."""
    net = SimpleNet(2, 4, 4)

    connect(
        net.cell(0).branch(0).loc(1.0),
        net.cell(1).branch(3).loc(1.0),
        IonotropicSynapse(),
    )
    connect(
        net.cell(1).branch(1).loc(0.8),
        net.cell(0).branch(3).loc(0.1),
        IonotropicSynapse(),
    )

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    net.cell(0).branch(0).loc(0.0).stimulate(current)
    net.cell(0).branch(0).loc(0.5).record()
    net.cell(1).branch(3).loc(0.5).record()

    voltages_jx_cpu = jx.integrate(net, voltage_solver="jaxley.dhs.cpu")
    voltages_jx_gpu = jx.integrate(net, voltage_solver="jaxley.dhs.gpu")
    voltages_jx_sparse = jx.integrate(net, voltage_solver="jax.sparse")

    message = "Voltages do not match between"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_gpu))
    assert max_error < 1e-8, f"{message} cpu/gpu. Error={max_error}"
    max_error = np.max(np.abs(voltages_jx_cpu - voltages_jx_sparse))
    assert max_error < 1e-8, f"{message} cpu/sparse. Error={max_error}"


def test_api_equivalence_synapses(SimpleNet):
    """Test whether ways of adding synapses are equivalent."""
    net1 = SimpleNet(2, 4, 4)

    connect(
        net1.cell(0).branch(0).loc(1.0),
        net1.cell(1).branch(3).loc(1.0),
        IonotropicSynapse(),
    )
    connect(
        net1.cell(1).branch(1).loc(0.8),
        net1.cell(0).branch(3).loc(0.1),
        IonotropicSynapse(),
    )

    net2 = SimpleNet(2, 4, 4)
    pre = net2.cell(0).branch(0).loc(1.0)
    post = net2.cell(1).branch(3).loc(1.0)
    connect(pre, post, IonotropicSynapse())

    pre = net2.cell(1).branch(1).loc(0.8)
    post = net2.cell(0).branch(3).loc(0.1)
    connect(pre, post, IonotropicSynapse())

    for net in [net1, net2]:
        current = jx.step_current(
            i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
        )
        net.cell(0).branch(0).loc(0.0).stimulate(current)
        net.cell(0).branch(0).loc(0.5).record()
        net.cell(1).branch(3).loc(0.5).record()

    voltages1 = jx.integrate(net1)
    voltages2 = jx.integrate(net2)

    assert (
        np.max(np.abs(voltages1 - voltages2)) < 1e-8
    ), "Voltages do not match between synapse APIs."


def test_api_equivalence_continued_simulation(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.insert(HH())
    cell[0, 1].record()

    v1 = jx.integrate(cell, t_max=4.0)
    v21, states = jx.integrate(cell, return_states=True, t_max=2.0)
    v22 = jx.integrate(cell, all_states=states, t_max=2.0)

    v2 = jnp.concatenate([v21, v22[:, 1:]], axis=1)
    assert np.max(np.abs(v1 - v2)) < 1e-8


def test_api_equivalence_network_matches_cell(SimpleBranch):
    """Test whether a network with w=0 synapses equals the individual cells.

    This runs an unequal number of compartments per branch."""
    dt = 0.025  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )

    branch1 = SimpleBranch(ncomp=1)
    branch2 = SimpleBranch(ncomp=2)
    branch3 = SimpleBranch(ncomp=3)
    cell1 = jx.Cell([branch1, branch2, branch3], parents=[-1, 0, 0])
    cell2 = jx.Cell([branch1, branch2], parents=[-1, 0])
    cell1.insert(HH())
    cell2.insert(HH())

    net = jx.Network([cell1, cell2])
    pre = net.cell(0).branch(2).comp(2)
    post = net.cell(1).branch(1).comp(1)
    connect(pre, post, IonotropicSynapse())
    net.IonotropicSynapse.edge("all").set("IonotropicSynapse_gS", 0.0)

    net.cell(0).branch(2).comp(2).stimulate(current)
    net.cell(0).branch(0).comp(0).record()

    net.cell(1).branch(1).comp(1).stimulate(current)
    net.cell(1).branch(0).comp(0).record()
    voltages_net = jx.integrate(net, delta_t=dt, voltage_solver="jaxley.dhs.cpu")

    cell1.branch(2).comp(2).stimulate(current)
    cell1.branch(0).comp(0).record()

    cell2.branch(1).comp(1).stimulate(current)
    cell2.branch(0).comp(0).record()
    voltages_cell1 = jx.integrate(cell1, delta_t=dt, voltage_solver="jaxley.dhs.cpu")
    voltages_cell2 = jx.integrate(cell2, delta_t=dt, voltage_solver="jaxley.dhs.cpu")
    voltages_cells = jnp.concatenate([voltages_cell1, voltages_cell2], axis=0)

    max_error = np.max(np.abs(voltages_net - voltages_cells))
    assert max_error < 1e-8, f"Error is {max_error}"


def test_api_init_step_to_integrate(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.insert(HH())
    cell[0, 1].record()

    # Internal integration function API
    delta_t = 0.025  # Default delta_t is 0.025
    v1 = jx.integrate(cell, t_max=4.0, delta_t=delta_t)

    # Flexibe init and step API
    init_fn, step_fn = build_init_and_step_fn(cell)

    params = cell.get_parameters()
    states, params = init_fn(params)
    step_fn_ = jax.jit(step_fn)
    rec_inds = cell.recordings.rec_index.to_numpy()
    rec_states = cell.recordings.state.to_numpy()

    steps = int(4.0 / delta_t)  # Steps to integrate
    recordings = [
        states[rec_state][rec_ind][None]
        for rec_state, rec_ind in zip(rec_states, rec_inds)
    ]
    externals = cell.externals
    for _ in range(steps):
        states = step_fn_(states, params, externals, delta_t=delta_t)
        recs = jnp.asarray(
            [
                states[rec_state][rec_ind]
                for rec_state, rec_ind in zip(rec_states, rec_inds)
            ]
        )
        recordings.append(recs)

    rec = jnp.stack(recordings, axis=0).T

    assert jnp.allclose(v1, rec)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import os

import jax.numpy as jnp
import numpy as np
from jax import jit

import jaxley as jx


def test_move_cell(SimpleBranch, SimpleCell):
    # Test move on a cell with compute_xyz()
    cell = SimpleCell(5, ncomp=4)
    cell.compute_xyz()
    cell.move(20.0, 30.0, 5.0)
    assert cell.xyzr[0][0, 0] == 20.0
    assert cell.xyzr[0][0, 1] == 30.0
    assert cell.xyzr[0][0, 2] == 5.0

    # Test move_to on a cell that starts with a specified xyzr
    branch = SimpleBranch(ncomp=4)
    cell = jx.Cell(
        branch,
        parents=[-1],
        xyzr=[
            np.array(
                [
                    [5.0, 10.0, 1.0, 10.0],
                    [10.0, 15.0, 0.0, 10.0],
                ]
            )
        ],
    )
    cell.move(6.0, 9.0, 3.0)
    assert cell.xyzr[0][0, 0] == 11.0
    assert cell.xyzr[0][0, 1] == 19.0
    assert cell.xyzr[0][0, 2] == 4.0
    assert cell.xyzr[0][0, 3] == 10.0


def test_move_network(SimpleCell):
    cell = SimpleCell(3, 3)
    cell.compute_xyz()
    net = jx.Network([cell, cell, cell])
    net.move(20.0, 30.0, 5.0)
    for i in [0, 3, 6]:
        assert net.xyzr[i][0, 0] == 20.0
        assert net.xyzr[i][0, 1] == 30.0
        assert net.xyzr[i][0, 2] == 5.0


def test_move_to_cell(SimpleBranch, SimpleCell):
    cell = SimpleCell(5, 4)
    cell.compute_xyz()
    cell.move_to(20.0, 30.0, 5.0)
    assert cell.xyzr[0][0, 0] == 20.0
    assert cell.xyzr[0][0, 1] == 30.0
    assert cell.xyzr[0][0, 2] == 5.0

    branch = SimpleBranch(ncomp=4)
    cell = jx.Cell(
        branch,
        parents=[-1],
        xyzr=[
            np.array(
                [
                    [5.0, 10.0, 1.0, 10.0],
                    [10.0, 15.0, 0.0, 10.0],
                ]
            )
        ],
    )
    cell.move_to(6.0, 9.0, 3.0)
    assert cell.xyzr[0][0, 0] == 6.0
    assert cell.xyzr[0][0, 1] == 9.0
    assert cell.xyzr[0][0, 2] == 3.0
    assert cell.xyzr[0][0, 3] == 10.0


def test_move_to_network(SimpleNet):
    net = SimpleNet(3, 3, 4)
    net.compute_xyz()
    net.move_to(10.0, 20.0, 30.0)
    # Branch 0 of cell 0
    assert net.xyzr[0][0, 0] == 10.0
    assert net.xyzr[0][0, 1] == 20.0
    assert net.xyzr[0][0, 2] == 30.0
    # Branch 0 of cell 1
    assert net.xyzr[3][0, 0] == 10.0
    assert net.xyzr[3][0, 1] == 20.0
    assert net.xyzr[3][0, 2] == 30.0


def test_move_to_arrays(SimpleNet):
    """Test with network"""
    ncomp = 4
    net = SimpleNet(3, 3, ncomp)
    net.compute_xyz()
    x_coords = np.array([10.0, 20.0, 30.0])
    y_coords = np.array([5.0, 15.0, 25.0])
    z_coords = np.array([1.0, 2.0, 3.0])
    net.move_to(x_coords, y_coords, z_coords)
    assert net.xyzr[0][0, 0] == 10.0
    assert net.xyzr[0][1, 0] == ncomp * 10.0 + 10.0
    assert net.xyzr[0][0, 1] == 5.0
    assert net.xyzr[0][0, 2] == 1.0
    assert net.xyzr[3][0, 0] == 20.0
    assert net.xyzr[3][0, 2] == 2.0
    assert net.xyzr[6][0, 1] == 25.0


def test_move_to_cellview(SimpleNet):
    net = SimpleNet(3, 3, 2)
    net.compute_xyz()

    # Test with float input
    net.cell(0).move_to(50.0, 3.0, 40.0)
    assert net.xyzr[0][0, 0] == 50.0
    assert net.xyzr[0][0, 1] == 3.0
    assert net.xyzr[0][0, 2] == 40.0

    # Test with array input
    net = SimpleNet(4, 3, 2)
    net.compute_xyz()
    testx = np.array([1.0, 2.0, 3.0])
    testy = np.array([4.0, 5.0, 6.0])
    testz = np.array([7.0, 8.0, 9.0])
    net.cell([0, 1, 2]).move_to(testx, testy, testz)
    assert net.xyzr[0][0, 0] == 1.0
    assert net.xyzr[3][0, 1] == 5.0
    assert net.xyzr[6][0, 2] == 9.0
    assert net.xyzr[9][0, 0] == 0.0


def test_move_to_swc_cell(SimpleMorphCell):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
    cell1 = SimpleMorphCell(fname, ncomp=1)
    cell2 = SimpleMorphCell(fname, ncomp=1)
    cell3 = SimpleMorphCell(fname, ncomp=1)

    # Try move_to on a cell
    cell1.move_to(10.0, 20.0, 30.0)
    assert cell1.xyzr[0][0, 0] == 10.0
    assert cell1.xyzr[0][0, 1] == 20.0

    net = jx.Network([cell1, cell2, cell3])
    # Try move_to on the network
    net.move_to(10.0, 20.0, 30.0)
    assert net.xyzr[0][0, 0] == 10.0
    assert net.xyzr[0][0, 1] == 20.0

    x_coords = np.array([20.0, 30.0, 40.0])
    y_coords = np.array([5.0, 15.0, 25.0])
    z_coords = np.array([1.0, 2.0, 3.0])

    net.move_to(x_coords, y_coords, z_coords)

    n_branches_cell1 = len(cell1.xyzr)
    n_branches_cell2 = len(cell2.xyzr)
    assert net.xyzr[0][0, 0] == 20.0
    assert net.xyzr[n_branches_cell1][0, 1] == 15.0
    assert net.xyzr[n_branches_cell1 + n_branches_cell2][0, 2] == 3.0

    # Test move_to on a subset of the cells
    sub_net = net.cell([1, 2])
    sub_net.move_to(30.0, 40.0, 50.0)
    assert net.xyzr[n_branches_cell1][0, 0] == 30.0
    assert net.xyzr[n_branches_cell1][0, 1] == 40.0

    x_coords = np.array([3.0, 4.0])
    y_coords = np.array([6.0, 7.0])
    z_coords = np.array([9.0, 10.0])
    sub_net.move_to(x_coords, y_coords, z_coords)
    assert net.xyzr[n_branches_cell1][0, 0] == 3.0
    assert net.xyzr[n_branches_cell1 + n_branches_cell2][0, 1] == 7.0



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax
import pytest

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np

import jaxley as jx
from jaxley.channels import HH
from jaxley.connect import fully_connect
from jaxley.synapses import IonotropicSynapse, TestSynapse


def test_record_and_stimulate_api(SimpleCell):
    """Test the API for recording and stimulating."""
    cell = SimpleCell(3, 2)

    cell.branch(0).loc(0.0).record()
    cell.branch(1).loc(1.0).record()

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    cell.branch(1).loc(1.0).stimulate(current)

    cell.delete_recordings()
    cell.delete_stimuli()


def test_record_shape(SimpleCell):
    """Test the API for recording and stimulating."""
    cell = SimpleCell(3, 2)

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    cell.branch(1).loc(1.0).stimulate(current)

    cell.branch(0).loc(0.0).record()
    cell.branch(1).loc(1.0).record()
    cell.branch(0).loc(1.0).record()
    cell.delete_recordings()
    cell.branch(2).loc(0.5).record()
    cell.branch(1).loc(0.1).record()

    voltages = jx.integrate(cell)
    assert (
        voltages.shape[0] == 2
    ), f"Shape of recordings ({voltages.shape}) is not right."


def test_record_synaptic_and_membrane_states(SimpleNet):
    """Tests recording of synaptic and membrane states.

    Tests are functional, not just API. They test whether the voltage and synaptic
    states spike at (almost) the same times.
    """

    _ = np.random.seed(0)  # Seed because connectivity is at random postsyn locs.

    net = SimpleNet(3, 1, 4)
    net.insert(HH())

    fully_connect(net.cell([0]), net.cell([1]), IonotropicSynapse())
    fully_connect(net.cell([1]), net.cell([2]), TestSynapse())
    fully_connect(net.cell([2]), net.cell([0]), IonotropicSynapse())

    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    net.cell(0).branch(0).loc(0.0).stimulate(current)

    # Invoke recording of voltage and synaptic states.
    net.cell(2).branch(0).loc(0.0).record("v")
    net.IonotropicSynapse.edge(1).record("IonotropicSynapse_s")
    net.cell(2).branch(0).loc(0.0).record("HH_m")
    net.cell(1).branch(0).loc(0.0).record("v")
    net.TestSynapse.edge(0).record("TestSynapse_c")
    net.cell(1).branch(0).loc(0.0).record("HH_m")
    net.cell(1).branch(0).loc(0.0).record("i_HH")
    net.IonotropicSynapse.edge(1).record("i_IonotropicSynapse")

    # Advanced synapse indexing for recording.
    net.copy_node_property_to_edges("global_cell_index")
    # Record currents from specific post synaptic cells.
    df = net.edges
    df = df.query("pre_global_cell_index in [0, 1]")
    net.select(edges=df.index).record("i_IonotropicSynapse")
    # Record currents from specific synapse types
    df = net.edges
    df = df.query("type == 'TestSynapse'")
    net.select(edges=df.index).record("i_TestSynapse")

    recs = jx.integrate(net)

    # Loop over first two recordings and then the second two recordings.
    for index in [0, 3]:
        # Local maxima of voltage trace.
        y = recs[index]
        maxima_1 = np.where((y[1:-1] > y[0:-2]) * (y[1:-1] > y[2:]))[0] + 1
        max_vals = recs[index][maxima_1]
        condition = max_vals > 10.0
        maxima_1 = maxima_1[condition]

        # Local maxima of synaptic state.
        y = recs[index + 1]
        maxima_2 = np.where((y[1:-1] > y[0:-2]) * (y[1:-1] > y[2:]))[0] + 1

        # Local maxima of membrane channel trace.
        y = recs[index + 2]
        maxima_3 = np.where((y[1:-1] > y[0:-2]) * (y[1:-1] > y[2:]))[0] + 1
        max_vals = recs[index + 2][maxima_3]
        condition = max_vals > 0.3
        maxima_3 = maxima_3[condition]

        # On average the synaptic trace spikes around 10 steps after voltage.
        offset_syn = 10
        assert np.all(np.abs(maxima_2 - maxima_1 - offset_syn)) < 5.0

        # On average the membrane trace spikes around 0 steps after voltage.
        offset_mem = 0
        assert np.all(np.abs(maxima_3 - maxima_1 - offset_mem)) < 5.0


def test_empty_recordings(SimpleComp):
    # Create an empty compartment
    comp = SimpleComp()

    # Check if a ValueError is raised when integrating an empty compartment
    with pytest.raises(ValueError):
        v = jx.integrate(comp, delta_t=0.025, t_max=10.0)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".4"

import jax.numpy as jnp
import numpy as np
from neuron import h

import jaxley as jx
from jaxley.channels import HH

_ = h.load_file("stdlib.hoc")
_ = h.load_file("import3d.hoc")


def test_similarity():
    """Test similarity of jaxley vs neuron."""
    i_delay = 3.0  # ms
    i_dur = 2.0  # ms
    i_amp = 1.0  # nA

    dt = 0.025  # ms
    t_max = 10.0  # ms

    voltages_jaxley = _run_jaxley(i_delay, i_dur, i_amp, dt, t_max)
    voltages_neuron = _run_neuron(i_delay, i_dur, i_amp, dt, t_max)

    assert np.mean(np.abs(voltages_jaxley - voltages_neuron)) < 0.05


def _run_jaxley(i_delay, i_dur, i_amp, dt, t_max):
    comp = jx.Compartment()
    comp.insert(HH())

    comp.set("length", 10.0)
    comp.set("radius", 10.0)

    comp.set("HH_gNa", 0.120)
    comp.set("HH_gK", 0.036)
    comp.set("HH_gLeak", 0.0003)
    comp.set("HH_m", 0.07490098835688629)
    comp.set("HH_h", 0.488947681848153)
    comp.set("HH_n", 0.3644787002343737)
    comp.set("v", -62.0)
    comp.set("capacitance", 5.0)

    current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    comp.stimulate(current)
    comp.record()

    voltages = jx.integrate(comp, delta_t=dt)
    return voltages


def _run_neuron(i_delay, i_dur, i_amp, dt, t_max):
    h.dt = dt
    h.secondorder = 0

    for sec in h.allsec():
        h.delete_section(sec=sec)

    comp = h.Section()

    comp.nseg = 1
    comp.Ra = 1_000.0
    comp.L = 10.0
    comp.diam = 20.0
    comp.cm = 5.0

    comp.insert("hh")
    comp.gnabar_hh = 0.120  # S/cm2
    comp.gkbar_hh = 0.036  # S/cm2
    comp.gl_hh = 0.0003  # S/cm2
    comp.ena = 50  # mV
    comp.ek = -77.0  # mV
    comp.el_hh = -54.3  # mV

    stim = h.IClamp(comp(0.0))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    voltage1 = h.Vector()
    voltage1.record(comp(0.0)._ref_v)

    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    voltages = np.asarray([list(voltage1)])
    return voltages



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import pytest
from jax import value_and_grad

import jaxley as jx
from jaxley.channels import HH


@pytest.mark.parametrize("key", ["HH_m", "v"])
def test_grad_against_finite_diff_initial_state(key, SimpleComp):
    def simulate():
        return jnp.sum(jx.integrate(comp))

    def simulate_with_params(params):
        return jnp.sum(jx.integrate(comp, params=params))

    comp = SimpleComp(copy=True)
    comp.insert(HH())
    comp.record()
    comp.stimulate(
        jx.step_current(i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0)
    )

    val = 0.2 if key == "HH_m" else -70.0
    step_size = 0.01

    # Finite differences.
    comp.set(key, val)
    r1 = simulate()
    comp.set(key, val + step_size)
    r2 = simulate()
    finitediff_grad = (r2 - r1) / step_size

    # Autodiff gradient.
    grad_fn = value_and_grad(simulate_with_params)
    comp.set(key, val)
    comp.make_trainable(key)
    params = comp.get_parameters()
    v, g = grad_fn(params)
    autodiff_grad = g[0][key]

    # Less than 5% error on the gradient difference.
    assert jnp.abs(autodiff_grad - finitediff_grad) / autodiff_grad < 0.05


@pytest.mark.parametrize("key", ["HH_m", "v"])
def test_branch_grad_against_finite_diff_initial_state(key, SimpleBranch):
    def simulate():
        return jnp.sum(jx.integrate(branch))

    def simulate_with_params(params):
        return jnp.sum(jx.integrate(branch, params=params))

    branch = SimpleBranch(4)
    branch.loc(0.0).record()
    branch.loc(0.0).stimulate(
        jx.step_current(i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0)
    )
    branch.loc(0.0).insert(HH())

    val = 0.2 if key == "HH_m" else -70.0
    step_size = 0.01

    # Finite differences.
    branch.loc(0.0).set(key, val)
    r1 = simulate()
    branch.loc(0.0).set(key, val + step_size)
    r2 = simulate()
    finitediff_grad = (r2 - r1) / step_size

    # Autodiff gradient.
    grad_fn = value_and_grad(simulate_with_params)
    branch.loc(0.0).set(key, val)
    branch.loc(0.0).make_trainable(key)
    params = branch.get_parameters()
    v, g = grad_fn(params)
    autodiff_grad = g[0][key]

    # Less than 6% error on the gradient difference, error for HH_m is around 5.2%.
    assert jnp.abs(autodiff_grad - finitediff_grad) / autodiff_grad < 0.06



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
import jax.numpy as jnp
from jax.typing import ArrayLike


def save_exp(x, max_value: float = 20.0):
    """Clip the input to a maximum value and return its exponential."""
    x = jnp.clip(x, a_max=max_value)
    return jnp.exp(x)


def solve_gate_implicit(
    gating_state: ArrayLike,
    dt: float,
    alpha: ArrayLike,
    beta: ArrayLike,
):
    a_m = gating_state + dt * alpha
    b_m = 1.0 + dt * alpha + dt * beta

    return a_m / b_m


def solve_gate_exponential(
    x: ArrayLike,
    dt: float,
    alpha: ArrayLike,
    beta: ArrayLike,
):
    tau = 1 / (alpha + beta)
    xinf = alpha * tau
    return exponential_euler(x, dt, xinf, tau)


def exponential_euler(
    x: ArrayLike,
    dt: float,
    x_inf: ArrayLike,
    x_tau: ArrayLike,
):
    """An exact solver for the linear dynamical system `dx = -(x - x_inf) / x_tau`."""
    exp_term = save_exp(-dt / x_tau)
    return x * exp_term + x_inf * (1.0 - exp_term)


def solve_inf_gate_exponential(
    x: ArrayLike,
    dt: float,
    s_inf: ArrayLike,
    tau_s: ArrayLike,
):
    """solves dx/dt = (s_inf - x) / tau_s
    via exponential Euler

    Args:
        x (ArrayLike): gate variable
        dt (float): time_delta
        s_inf (ArrayLike): _description_
        tau_s (ArrayLike): _description_

    Returns:
        _type_: updated gate
    """
    slope = -1.0 / tau_s
    exp_term = save_exp(slope * dt)
    return x * exp_term + s_inf * (1.0 - exp_term)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import List, Optional

from jaxley.io.graph import build_compartment_graph, from_graph, to_swc_graph
from jaxley.modules import Cell


def read_swc(
    fname: str,
    ncomp: Optional[int],
    max_branch_len: Optional[float] = None,
    min_radius: Optional[float] = None,
    assign_groups: bool = True,
    backend: str = "graph",
    ignore_swc_tracing_interruptions: bool = True,
    relevant_type_ids: Optional[List[int]] = None,
) -> Cell:
    """Reads SWC file into a `Cell`.

    Jaxley assumes cylindrical compartments and therefore defines length and radius
    for every compartment. The surface area is then 2*pi*r*length. For branches
    consisting of a single traced point we assume for them to have area 4*pi*r*r.
    Therefore, in these cases, we set length=2*r.

    Args:
        fname: Path to the swc file.
        ncomp: The number of compartments per branch.
        max_branch_len: If a branch is longer than this value it is split into two
            branches.
        min_radius: If the radius of a reconstruction is below this value it is clipped.
        assign_groups: If True, then the identity of reconstructed points in the SWC
            file will be used to generate groups `soma`, `axon`, `basal`, `apical`. See
            here:
            http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html
        backend: The backend to use. Currently only `graph` is supported.
        ignore_swc_tracing_interruptions: Whether to ignore discontinuities in the swc
            tracing order. If False, this will result in split branches at these points.
        relevant_type_ids: All type ids that are not in this list will be ignored for
            tracing the morphology. This means that branches which have multiple type
            ids (which are not in `relevant_type_ids`) will be considered as one branch.
            If `None`, we default to `[1, 2, 3, 4]`.

    Returns:
        A `Cell` object."""

    if backend == "graph":
        swc_graph = to_swc_graph(fname)
        comp_graph = build_compartment_graph(
            swc_graph,
            ncomp=ncomp,
            root=None,
            min_radius=min_radius,
            max_len=max_branch_len,
            ignore_swc_tracing_interruptions=ignore_swc_tracing_interruptions,
            relevant_type_ids=relevant_type_ids,
        )
        module = from_graph(
            comp_graph,
            assign_groups=assign_groups,
            solve_root=None,
            traverse_for_solve_order=True,  # Traverse to fix potential tracing errors.
        )
        return module
    else:
        raise ValueError(f"Unknown backend: {backend}. Use either `custom` or `graph`.")



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

# This folder that contains tests which ensure that the results are almost identical
# to the results of the neuron toolbox.
